# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from functools import reduce
from mrmustard.utils import training
from mrmustard.types import *
from mrmustard.math import Math

math = Math()


class Parametrized:
    r"""Abstract base class for all parametrized objects (gates, detectors, etc.)

    For each trainable parameter keyword arguments must be passed for the initial value ``xxx``
    (tensor), the numerical bounds ``xxx_bounds`` (float, float), whether the parameter ``xxx`` will
    be optimized ``xxx_trainable`` (bool), along with any other parameters.
    """

    def __init__(self, **kwargs):  # NOTE: only kwargs so that we can use the arg names
        self._trainable_parameters = []
        self._constant_parameters = []
        self._param_names = []
        owner = f"{self.__class__.__qualname__}"
        for name, value in kwargs.items():
            if math.from_backend(value):
                if math.is_trainable(value):
                    self._trainable_parameters.append(value)
                elif name + "_trainable" in kwargs and kwargs[name + "_trainable"]:
                    value = math.new_variable(value, kwargs[name + "_bounds"], owner + ":" + name)
                    self._trainable_parameters.append(value)
                else:
                    self._constant_parameters.append(value)
            elif name + "_trainable" in kwargs and kwargs[name + "_trainable"]:
                value = math.new_variable(value, kwargs[name + "_bounds"], owner + ":" + name)
                self._trainable_parameters.append(value)
            elif name + "_trainable" in kwargs and not kwargs[name + "_trainable"]:
                value = math.new_constant(value, owner + ":" + name)
                self._constant_parameters.append(value)
            else:
                name = "_" + name
            self.__dict__[name] = value
            self._param_names += [] if name.startswith("_") else [name]

    @property
    def trainable_parameters(self) -> Dict[str, List[Trainable]]:
        r"""Returns the dictionary of trainable parameters, searching recursively in the object tree (for example, when in a Circuit)."""
        if hasattr(self, "_ops"):
            return {
                "symplectic": math.unique_tensors(
                    [p for item in self._ops for p in item.trainable_parameters["symplectic"]]
                ),
                "orthogonal": math.unique_tensors(
                    [p for item in self._ops for p in item.trainable_parameters["orthogonal"]]
                ),
                "euclidean": math.unique_tensors(
                    [p for item in self._ops for p in item.trainable_parameters["euclidean"]]
                ),
            }
        else:
            return {
                "symplectic": [],
                "orthogonal": [],
                "euclidean": self._trainable_parameters,
            }  # default

    @property
    def constant_parameters(self) -> Dict[str, List[Tensor]]:
        r"""Returns the dictionary of constant parameters, searching recursively in the object tree (for example, when in a Circuit)."""
        if hasattr(self, "_ops"):
            return {
                "symplectic": math.unique_tensors(
                    [p for item in self._ops for p in item.constant_parameters["symplectic"]]
                ),
                "orthogonal": math.unique_tensors(
                    [p for item in self._ops for p in item.constant_parameters["orthogonal"]]
                ),
                "euclidean": math.unique_tensors(
                    [p for item in self._ops for p in item.constant_parameters["euclidean"]]
                ),
            }
        else:
            return {
                "symplectic": [],
                "orthogonal": [],
                "euclidean": self._constant_parameters,
            }  # default



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import tensorflow as tf
from thewalrus import hermite_multidimensional, grad_hermite_multidimensional

from .math_interface import MathInterface
from mrmustard.math.autocast import Autocast
from mrmustard.types import *


class TFMath(MathInterface):

    float64 = tf.float64
    float32 = tf.float32
    complex64 = tf.complex64
    complex128 = tf.complex128

    def __getattr__(self, name):
        return getattr(tf, name)

    # ~~~~~~~~~
    # Basic ops
    # ~~~~~~~~~

    def abs(self, array: tf.Tensor) -> tf.Tensor:
        return tf.abs(array)

    def any(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.reduce_any(array)

    def arange(self, start: int, limit: int = None, delta: int = 1, dtype=tf.float64) -> tf.Tensor:
        return tf.range(start, limit, delta, dtype=dtype)

    def asnumpy(self, tensor: tf.Tensor) -> Tensor:
        return np.array(tensor)

    def assign(self, array: tf.Tensor, value: tf.Tensor) -> tf.Tensor:
        array.assign(value)
        return array

    def astensor(self, array: Union[np.ndarray, tf.Tensor], dtype=None) -> tf.Tensor:
        return tf.convert_to_tensor(array, dtype=dtype)

    def atleast_1d(self, array: tf.Tensor, dtype=None) -> tf.Tensor:
        return self.cast(tf.reshape(array, [-1]), dtype)

    def cast(self, x: tf.Tensor, dtype=None) -> tf.Tensor:
        if dtype is None:
            return x
        return tf.cast(x, dtype)

    def clip(self, array, a_min, a_max) -> tf.Tensor:
        return tf.clip_by_value(array, a_min, a_max)

    def concat(self, values: Sequence[tf.Tensor], axis: int) -> tf.Tensor:
        return tf.concat(values, axis)

    def conj(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.conj(array)

    def constraint_func(
        self, bounds: Tuple[Optional[float], Optional[float]]
    ) -> Optional[Callable]:
        bounds = (
            -np.inf if bounds[0] is None else bounds[0],
            np.inf if bounds[1] is None else bounds[1],
        )
        if not bounds == (-np.inf, np.inf):
            constraint: Optional[Callable] = lambda x: tf.clip_by_value(x, bounds[0], bounds[1])
        else:
            constraint = None
        return constraint

    @Autocast()
    def convolution(
        self,
        array: tf.Tensor,
        filters: tf.Tensor,
        strides: Optional[List[int]] = None,
        padding="VALID",
        data_format="NWC",
        dilations: Optional[List[int]] = None,
    ) -> tf.Tensor:
        return tf.nn.convolution(array, filters, strides, padding, data_format, dilations)

    def cos(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.cos(array)

    def cosh(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.cosh(array)

    def det(self, a: tf.Tensor) -> tf.Tensor:
        return tf.linalg.det(a)

    def diag(self, array: tf.Tensor, k: int = 0) -> tf.Tensor:
        return tf.linalg.diag(array, k=k)

    def diag_part(self, array: tf.Tensor) -> tf.Tensor:
        return tf.linalg.diag_part(array)

    def einsum(self, string: str, *tensors) -> tf.Tensor:
        return tf.einsum(string, *tensors)

    def exp(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.exp(array)

    def expand_dims(self, array: tf.Tensor, axis: int) -> tf.Tensor:
        return tf.expand_dims(array, axis)

    def expm(self, matrix: tf.Tensor) -> tf.Tensor:
        return tf.linalg.expm(matrix)

    def eye(self, size: int, dtype=tf.float64) -> tf.Tensor:
        return tf.eye(size, dtype=dtype)

    def from_backend(self, thing) -> bool:
        return isinstance(thing, (tf.Tensor, tf.Variable))

    def gather(self, array: tf.Tensor, indices: tf.Tensor, axis: int = None) -> tf.Tensor:
        return tf.gather(array, indices, axis=axis)

    def hash_tensor(self, tensor: tf.Tensor) -> int:
        try:
            REF = tensor.ref()
        except AttributeError:
            raise TypeError(f"Cannot hash tensor")
        return hash(REF)

    def imag(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.imag(array)

    def inv(self, a: tf.Tensor) -> tf.Tensor:
        return tf.linalg.inv(a)

    def is_trainable(self, tensor: tf.Tensor) -> bool:
        return isinstance(tensor, tf.Variable)

    def lgamma(self, x: tf.Tensor) -> tf.Tensor:
        return tf.math.lgamma(x)

    def log(self, x: tf.Tensor) -> tf.Tensor:
        return tf.math.log(x)

    @Autocast()
    def matmul(
        self,
        a: tf.Tensor,
        b: tf.Tensor,
        transpose_a=False,
        transpose_b=False,
        adjoint_a=False,
        adjoint_b=False,
    ) -> tf.Tensor:
        return tf.linalg.matmul(a, b, transpose_a, transpose_b, adjoint_a, adjoint_b)

    @Autocast()
    def matvec(self, a: tf.Tensor, b: tf.Tensor, transpose_a=False, adjoint_a=False) -> tf.Tensor:
        return tf.linalg.matvec(a, b, transpose_a, adjoint_a)

    @Autocast()
    def maximum(self, a: tf.Tensor, b: tf.Tensor) -> tf.Tensor:
        return tf.maximum(a, b)

    @Autocast()
    def minimum(self, a: tf.Tensor, b: tf.Tensor) -> tf.Tensor:
        return tf.minimum(a, b)

    def new_variable(
        self, value, bounds: Tuple[Optional[float], Optional[float]], name: str, dtype=tf.float64
    ):
        value = self.convert_to_tensor(value, dtype)
        return tf.Variable(value, name=name, dtype=dtype, constraint=self.constraint_func(bounds))

    def new_constant(self, value, name: str, dtype=tf.float64):
        value = self.convert_to_tensor(value, dtype)
        return tf.constant(value, dtype=dtype, name=name)

    def norm(self, array: tf.Tensor) -> tf.Tensor:
        """Note that the norm preserves the type of array."""
        return tf.linalg.norm(array)

    def ones(self, shape: Sequence[int], dtype=tf.float64) -> tf.Tensor:
        return tf.ones(shape, dtype=dtype)

    def ones_like(self, array: tf.Tensor) -> tf.Tensor:
        return tf.ones_like(array)

    @Autocast()
    def outer(self, array1: tf.Tensor, array2: tf.Tensor) -> tf.Tensor:
        return tf.tensordot(array1, array2, [[], []])

    def pad(
        self,
        array: tf.Tensor,
        paddings: Sequence[Tuple[int, int]],
        mode="CONSTANT",
        constant_values=0,
    ) -> tf.Tensor:
        return tf.pad(array, paddings, mode, constant_values)

    def pinv(self, array: tf.Tensor) -> tf.Tensor:
        return tf.linalg.pinv(array)

    @Autocast()
    def pow(self, array: tf.Tensor, exponent: float) -> tf.Tensor:
        return tf.math.pow(array, exponent)

    def real(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.real(array)

    def reshape(self, array: tf.Tensor, shape: Sequence[int]) -> tf.Tensor:
        return tf.reshape(array, shape)

    def sin(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.sin(array)

    def sinh(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.sinh(array)

    def sqrt(self, x: tf.Tensor, dtype=None) -> tf.Tensor:
        return tf.sqrt(self.cast(x, dtype))

    def sum(self, array: tf.Tensor, axes: Sequence[int] = None):
        return tf.reduce_sum(array, axes)

    @Autocast()
    def tensordot(self, a: tf.Tensor, b: tf.Tensor, axes: List[int]) -> tf.Tensor:
        return tf.tensordot(a, b, axes)

    def tile(self, array: tf.Tensor, repeats: Sequence[int]) -> tf.Tensor:
        return tf.tile(array, repeats)

    def trace(self, array: tf.Tensor, dtype=None) -> tf.Tensor:
        return self.cast(tf.linalg.trace(array), dtype)

    def transpose(self, a: tf.Tensor, perm: Sequence[int] = None) -> tf.Tensor:
        if a is None:
            return None  # TODO: remove and address None inputs where tranpose is used
        return tf.transpose(a, perm)

    @Autocast()
    def update_tensor(self, tensor: tf.Tensor, indices: tf.Tensor, values: tf.Tensor):
        return tf.tensor_scatter_nd_update(tensor, indices, values)

    @Autocast()
    def update_add_tensor(self, tensor: tf.Tensor, indices: tf.Tensor, values: tf.Tensor):
        return tf.tensor_scatter_nd_add(tensor, indices, values)

    def unique_tensors(self, lst: List[Tensor]) -> List[Tensor]:
        hash_dict = dict()
        for tensor in lst:
            try:
                if (hash := self.hash_tensor(tensor)) not in hash_dict:
                    hash_dict[hash] = tensor
            except TypeError:
                continue
        return list(hash_dict.values())

    def zeros(self, shape: Sequence[int], dtype=tf.float64) -> tf.Tensor:
        return tf.zeros(shape, dtype=dtype)

    def zeros_like(self, array: tf.Tensor) -> tf.Tensor:
        return tf.zeros_like(array)

    # ~~~~~~~~~~~~~~~~~
    # Special functions
    # ~~~~~~~~~~~~~~~~~

    def DefaultEuclideanOptimizer(
        self,
    ) -> tf.keras.optimizers.Optimizer:  # TODO: a wrapper class is better?
        r"""Default optimizer for the Euclidean parameters."""
        return tf.keras.optimizers.Adam(learning_rate=0.001)

    def value_and_gradients(
        self, cost_fn: Callable, parameters: Dict[str, List[Trainable]]
    ) -> Tuple[tf.Tensor, Dict[str, List[tf.Tensor]]]:
        r"""Computes the loss and gradients of the given cost function.

        Args:
            cost_fn (Callable with no args): The cost function.
            parameters (Dict): The parameters to optimize in three kinds:
                symplectic, orthogonal and euclidean.

        Returns:
            The loss and the gradients.
        """
        with tf.GradientTape() as tape:
            loss = cost_fn()
        gradients = tape.gradient(loss, list(parameters.values()))
        return loss, {p: g for p, g in zip(parameters.keys(), gradients)}

    @tf.custom_gradient
    def hermite_renormalized(
        self, A: tf.Tensor, B: tf.Tensor, C: tf.Tensor, shape: Tuple[int]
    ) -> tf.Tensor:  # TODO this is not ready
        r"""Renormalized multidimensional Hermite polynomial given by the "exponential" Taylor
        series of :math:`exp(C + Bx - Ax^2)` at zero, where the series has :math:`sqrt(n!)` at the
        denominator rather than :math:`n!`. Note the minus sign in front of ``A``.

        Args:
            A: The A matrix.
            B: The B vector.
            C: The C scalar.
            shape: The shape of the final tensor.

        Returns:
            The renormalized Hermite polynomial of given shape.
        """
        poly = tf.numpy_function(
            hermite_multidimensional, [A, shape, B, C, True, True, True], A.dtype
        )

        def grad(dLdpoly):
            dpoly_dC, dpoly_dA, dpoly_dB = tf.numpy_function(
                grad_hermite_multidimensional, [poly, A, B, C], [poly.dtype] * 3
            )
            ax = tuple(range(dLdpoly.ndim))
            dLdA = self.sum(dLdpoly[..., None, None] * self.conj(dpoly_dA), axes=ax)
            dLdB = self.sum(dLdpoly[..., None] * self.conj(dpoly_dB), axes=ax)
            dLdC = self.sum(dLdpoly * self.conj(dpoly_dC), axes=ax)
            return dLdA, dLdB, dLdC

        return poly, grad

    def eigvals(self, tensor: tf.Tensor) -> Tensor:
        """Returns the eigenvalues of a matrix."""
        return tf.linalg.eigvals(tensor)

    def eigvalsh(self, tensor: tf.Tensor) -> Tensor:
        """Returns the eigenvalues of a Real Symmetric or Hermitian matrix."""
        return tf.linalg.eigvalsh(tensor)

    def svd(self, tensor: tf.Tensor) -> Tensor:
        """Returns the Singular Value Decomposition of a matrix."""
        return tf.linalg.svd(tensor)

    def xlogy(self, x: tf.Tensor, y: tf.Tensor) -> Tensor:
        """Returns 0 if ``x == 0,`` and ``x * log(y)`` otherwise, elementwise."""
        return tf.math.xlogy(x, y)

    def eigh(self, tensor: tf.Tensor) -> Tensor:
        """Returns the eigenvalues and eigenvectors of a matrix."""
        return tf.linalg.eigh(tensor)

    def sqrtm(self, tensor: tf.Tensor, rtol=1e-05, atol=1e-08) -> Tensor:
        """Returns the matrix square root of a square matrix, such that ``sqrt(A) @ sqrt(A) = A``."""

        # The sqrtm function has issues with matrices that are close to zero, hence we branch
        if np.allclose(tensor, 0, rtol=rtol, atol=atol):
            return self.zeros_like(tensor)
        else:
            return tf.linalg.sqrtm(tensor)

    def boolean_mask(self, tensor: tf.Tensor, mask: tf.Tensor) -> Tensor:
        """Returns a tensor based on the truth value of the boolean mask."""
        return tf.boolean_mask(tensor, mask)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Extras (not in the Interface)
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    @tf.custom_gradient
    def getitem(tensor, *, key):
        """A differentiable pure equivalent of numpy's ``value = tensor[key]``."""
        value = np.array(tensor)[key]

        def grad(dy):
            dL_dtensor = np.zeros_like(tensor)
            dL_dtensor[key] = dy
            return dL_dtensor

        return value, grad

    @tf.custom_gradient
    def setitem(tensor, value, *, key):
        """A differentiable pure equivalent of numpy's ``tensor[key] = value``."""
        tensor = np.array(tensor)
        value = np.array(value)
        tensor[key] = value

        def grad(dy):
            dL_dtensor = np.array(dy)
            dL_dtensor[key] = 0.0
            # unbroadcasting the gradient
            implicit_broadcast = list(range(tensor.ndim - value.ndim))
            explicit_broadcast = [
                tensor.ndim - value.ndim + j for j in range(value.ndim) if value.shape[j] == 1
            ]
            dL_dvalue = np.sum(
                np.array(dy)[key], axis=tuple(implicit_broadcast + explicit_broadcast)
            )
            dL_dvalue = np.expand_dims(
                dL_dvalue, [i - len(implicit_broadcast) for i in explicit_broadcast]
            )
            return dL_dtensor, dL_dvalue

        return tensor, grad



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from abc import ABC, abstractmethod
from mrmustard.math import Math

math = Math()
from mrmustard.physics import fock
from mrmustard.lab.abstract.state import State
from mrmustard.types import *
from mrmustard.utils import graphics
from mrmustard import settings
import numpy as np


class FockMeasurement(ABC):
    r"""A Fock measurement projecting onto a Fock measurement pattern.

    It works by representing the state in the Fock basis and then applying a stochastic channel
    matrix ``P(meas|n)`` to the Fock probabilities (belief propagation).

    It outputs the measurement probabilities and the remaining post-measurement state (if any)
    in the Fock basis.
    """

    def primal(self, state: State) -> Tensor:
        r"""
        Returns a tensor representing the post-measurement state in the unmeasured modes in the Fock basis.
        The first N indices of the returned tensor correspond to the Fock measurements of the N modes that
        the detector is measuring. The remaining indices correspond to the density matrix of the unmeasured modes.
        """
        cutoffs = []
        used = 0
        for mode in state.modes:
            if mode in self._modes:
                cutoffs.append(
                    max(settings.PNR_INTERNAL_CUTOFF, state.cutoffs[state.indices(mode)])
                )
                used += 1
            else:
                cutoffs.append(state.cutoffs[state.indices(mode)])
        if self.should_recompute_stochastic_channel() or math.any(
            [c > settings.PNR_INTERNAL_CUTOFF for c in state.cutoffs]
        ):
            self.recompute_stochastic_channel(cutoffs)
        dm = state.dm(cutoffs)
        for k, (mode, stoch) in enumerate(zip(self._modes, self._internal_stochastic_channel)):
            # move the mode indices to the end
            last = [mode - k, mode + state.num_modes - 2 * k]
            perm = [m for m in range(dm.ndim) if m not in last] + last
            dm = math.transpose(dm, perm)
            # compute sum_m P(meas|m)rho_mm
            dm = math.diag_part(dm)
            dm = math.tensordot(dm, stoch[: self._cutoffs[k], : dm.shape[-1]], [[-1], [1]])
        # put back the last len(self.modes) modes at the beginning
        output = math.transpose(
            dm,
            list(range(dm.ndim - len(self._modes), dm.ndim))
            + list(range(dm.ndim - len(self._modes))),
        )
        if len(output.shape) == len(self._modes):  # all modes are measured
            output = math.real(output)  # return probabilities
        return output

    def should_recompute_stochastic_channel(self) -> bool:  # override in subclasses
        return False

    def __lshift__(self, other) -> Tensor:
        if isinstance(other, State):
            self.primal(other)
        else:
            raise TypeError(
                f"unsupported operand type(s) '{type(self).__name__}' << '{type(other).__name__}'"
            )

    def __getitem__(self, items) -> Callable:
        r"""
        Allows measurements to be used as ``output = meas[0,1](input)``, e.g. measuring modes 0
        and 1.
        """

        if isinstance(items, int):
            modes = [items]
        elif isinstance(items, slice):
            modes = list(range(items.start, items.stop, items.step))
        elif isinstance(items, (Sequence, Iterable)):
            modes = list(items)
        else:
            raise ValueError(f"{items} is not a valid slice or list of modes.")
        self.modes = modes
        return self



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from hypothesis import given, strategies as st
from hypothesis.extra.numpy import arrays
from mrmustard.lab import *


angle = st.floats(min_value=0, max_value=2 * np.pi)
positive = st.floats(min_value=0, allow_infinity=False, allow_nan=False)
real = st.floats(allow_infinity=False, allow_nan=False)
r = st.floats(
    min_value=0, max_value=0.5, allow_infinity=False, allow_nan=False
)  # reasonable squeezing magnitude
real_not_zero = st.one_of(st.floats(max_value=-0.00001), st.floats(min_value=0.00001))
integer = st.integers(min_value=0, max_value=2 ** 32 - 1)
small_float = st.floats(min_value=-0.1, max_value=0.1, allow_infinity=False, allow_nan=False)
medium_float = st.floats(min_value=-1.0, max_value=1.0, allow_infinity=False, allow_nan=False)
large_float = st.floats(min_value=-10.0, max_value=10.0, allow_infinity=False, allow_nan=False)
num_modes = st.integers(min_value=0, max_value=10)


@st.composite
def vector(draw, length):
    return draw(
        st.lists(st.floats(min_value=-1.0, max_value=1.0), min_size=length, max_size=length)
    )


# a strategy to produce a list of integers of length num_modes. the integers are all different and between 0 and num_modes
@st.composite
def modes(draw, num_modes):
    return draw(
        st.lists(
            st.integers(min_value=0, max_value=num_modes), min_size=num_modes, max_size=num_modes
        ).filter(lambda x: len(set(x)) == len(x))
    )


def array_of_(strategy, minlen=0, maxlen=None):
    return arrays(dtype=np.float64, shape=(st.integers(minlen, maxlen),), elements=strategy)


def none_or_(strategy):
    return st.one_of(st.just(None), strategy)


angle_bounds = st.tuples(none_or_(angle), none_or_(angle)).filter(
    lambda t: t[0] < t[1] if t[0] is not None and t[1] is not None else True
)
positive_bounds = st.tuples(none_or_(positive), none_or_(positive)).filter(
    lambda t: t[0] < t[1] if t[0] is not None and t[1] is not None else True
)
real_bounds = st.tuples(none_or_(real), none_or_(real)).filter(
    lambda t: t[0] < t[1] if t[0] is not None and t[1] is not None else True
)


@st.composite
def random_Rgate(draw, num_modes=None, trainable=False):
    return Rgate(
        angle=draw(angle),
        angle_bounds=draw(angle_bounds),
        angle_trainable=trainable,
    )


@st.composite
def random_Sgate(draw, num_modes=None, trainable=False, small=False):
    return Sgate(
        r=np.abs(draw(small_float)) if small else draw(r),
        phi=draw(angle),
        r_bounds=draw(positive_bounds),
        phi_bounds=draw(angle_bounds),
        r_trainable=trainable,
        phi_trainable=trainable,
    )


@st.composite
def random_Dgate(draw, num_modes=None, trainable=False, small=False):
    if small:
        x = draw(small_float)
        y = draw(small_float)
    else:
        x = draw(medium_float)
        y = draw(medium_float)
    return Dgate(
        x=x,
        y=y,
        x_bounds=draw(real_bounds),
        y_bounds=draw(real_bounds),
        x_trainable=trainable,
        y_trainable=trainable,
    )


@st.composite
def random_S2gate(draw, trainable=False):
    return S2gate(
        r=draw(r),
        phi=draw(angle),
        r_bounds=draw(positive_bounds),
        phi_bounds=draw(angle_bounds),
        r_trainable=trainable,
        phi_trainable=trainable,
    )


@st.composite
def random_BSgate(draw, trainable=False):
    return BSgate(
        theta=draw(angle),
        phi=draw(angle),
        theta_bounds=draw(angle_bounds),
        phi_bounds=draw(angle_bounds),
        theta_trainable=trainable,
        phi_trainable=trainable,
    )


@st.composite
def random_MZgate(draw, trainable=False):
    return MZgate(
        phi_a=draw(angle),
        phi_b=draw(angle),
        phi_a_bounds=draw(angle_bounds),
        phi_b_bounds=draw(angle_bounds),
        phi_a_trainable=trainable,
        phi_b_trainable=trainable,
        internal=draw(st.booleans()),
    )


@st.composite
def random_Interferometer(draw, num_modes, trainable=False):
    return Interferometer(num_modes=num_modes, orthogonal_trainable=trainable)


@st.composite
def random_Ggate(draw, num_modes, trainable=False):
    displacement = vector(2 * num_modes)
    return Ggate(
        num_modes=num_modes,
        displacement=draw(displacement),
        displacement_trainable=trainable,
    )


@st.composite
def single_mode_unitary(draw, small=False):
    return draw(
        st.one_of(random_Rgate(1), random_Sgate(1, small=small), random_Dgate(1, small=small))
    )


@st.composite
def two_mode_gate(draw):
    return draw(
        st.one_of(
            random_S2gate(),
            random_BSgate(),
            random_MZgate(),
            random_Ggate(num_modes=2),
            random_Interferometer(num_modes=2),
        )
    )


@st.composite
def n_mode_gate(draw, num_modes=None):
    return draw(st.one_of(random_Interferometer(num_modes), random_Ggate(num_modes)))


## states
@st.composite
def squeezed_vacuum(draw, num_modes):
    r = array_of_(r, num_modes, num_modes)
    phi = array_of_(angle, num_modes, num_modes)
    return SqueezedVacuum(r=draw(r), phi=draw(phi))


@st.composite
def displacedsqueezed(draw, num_modes):
    r = array_of_(small_float.filter(lambda r: r > 0.0), num_modes, num_modes)
    phi_ = array_of_(angle, num_modes, num_modes)
    x = array_of_(medium_float, num_modes, num_modes)
    y = array_of_(medium_float, num_modes, num_modes)
    return DisplacedSqueezed(r=draw(r), phi=draw(phi), x=draw(x), y=draw(x))


@st.composite
def coherent(draw, num_modes):
    x = array_of_(medium_float, num_modes, num_modes)
    y = array_of_(medium_float, num_modes, num_modes)
    return Coherent(x=draw(x), y=draw(y))


@st.composite
def tmsv(draw):
    r = array_of_(medium_float.filter(lambda r: r > 0.0), 2, 2)
    phi = array_of_(angle, 2, 2)
    return TMSV(r=draw(r), phi=draw(phi))


@st.composite
def thermal(draw, num_modes):
    n_mean = array_of_(medium_float.filter(lambda r: r > 0.0), num_modes, num_modes)
    return Thermal(n_mean=draw(n_mean))


@st.composite
def default_state(draw, num_modes):
    return draw(
        st.one_of(
            squeezed_vacuum(num_modes),
            displacedsqueezed(num_modes),
            coherent(num_modes),
            tmsv(num_modes),
            thermal(num_modes),
        )
    )


@st.composite
def default_pure_state(draw, num_modes):
    return draw(
        st.one_of(
            squeezed_vacuum(num_modes),
            displacedsqueezed(num_modes),
            coherent(num_modes),
            tmsv(num_modes),
        )
    )


@st.composite
def pure_state(draw, num_modes=1, small=False):
    S = draw(random_Sgate(num_modes, small=small))
    I = draw(random_Interferometer(num_modes))
    D = draw(random_Dgate(num_modes, small=small))
    return Vacuum(num_modes) >> S >> I >> D



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from scipy.linalg import expm
from mrmustard.types import *
from mrmustard.utils import graphics
from mrmustard.math import Math

math = Math()


class Optimizer:
    r"""An optimizer for any parametrized object: it can optimize euclidean, orthogonal and symplectic parameters.

    .. note::

        In the future it will also include a compiler, so that it will be possible to
        simplify the circuit/detector/gate/etc before the optimization and also
        compile other types of structures like error correcting codes and encoders/decoders.
    """

    def __init__(
        self, symplectic_lr: float = 0.1, orthogonal_lr: float = 0.1, euclidean_lr: float = 0.001
    ):
        self.symplectic_lr: float = symplectic_lr
        self.orthogonal_lr: float = orthogonal_lr
        self.euclidean_lr: float = euclidean_lr
        self.opt_history: List[float] = [0]

    def minimize(
        self, cost_fn: Callable, by_optimizing: Sequence[Trainable], max_steps: int = 1000
    ):
        r"""Minimizes the given cost function by optimizing circuits and/or detectors.

        Args:
            cost_fn (Callable): a function that will be executed in a differentiable context in
                order to compute gradients as needed
            by_optimizing (list of circuits and/or detectors and/or gates): a list of elements that
                contain the parameters to optimize
            max_steps (int): the minimization keeps going until the loss is stable or max_steps are
                reached (if ``max_steps=0`` it will only stop when the loss is stable)
        """
        try:
            params = {
                "symplectic": math.unique_tensors(
                    [p for item in by_optimizing for p in item.trainable_parameters["symplectic"]]
                ),
                "orthogonal": math.unique_tensors(
                    [p for item in by_optimizing for p in item.trainable_parameters["orthogonal"]]
                ),
                "euclidean": math.unique_tensors(
                    [p for item in by_optimizing for p in item.trainable_parameters["euclidean"]]
                ),
            }
            bar = graphics.Progressbar(max_steps)
            with bar:
                while not self.should_stop(max_steps):
                    cost, grads = math.value_and_gradients(cost_fn, params)
                    update_symplectic(params["symplectic"], grads["symplectic"], self.symplectic_lr)
                    update_orthogonal(params["orthogonal"], grads["orthogonal"], self.orthogonal_lr)
                    update_euclidean(params["euclidean"], grads["euclidean"], self.euclidean_lr)
                    self.opt_history.append(cost)
                    bar.step(math.asnumpy(cost))
        except KeyboardInterrupt:  # graceful exit
            return

    def should_stop(self, max_steps: int) -> bool:
        r"""Returns ``True`` if the optimization should stop (either because the loss is stable or because the maximum number of steps is reached)."""
        if max_steps != 0 and len(self.opt_history) > max_steps:
            return True
        if len(self.opt_history) > 20:  # if cost varies less than 10e-6 over 20 steps
            if (
                sum(abs(self.opt_history[-i - 1] - self.opt_history[-i]) for i in range(1, 20))
                < 1e-6
            ):
                print("Loss looks stable, stopping here.")
                return True
        return False


# ~~~~~~~~~~~~~~~~~
# Static functions
# ~~~~~~~~~~~~~~~~~


# def new_variable(
#     value, bounds: Tuple[Optional[float], Optional[float]], name: str, dtype=math.float64
# ) -> Trainable:
#     r"""Returns a new trainable variable from the current math backend
#     with initial value set by `value` and bounds set by `bounds`.
#
#     Args:
#         value (float): The initial value of the variable
#         bounds (Tuple[float, float]): The bounds of the variable
#         name (str): The name of the variable
#         dtype: The dtype of the variable
#
#     Returns:
#         variable (Trainable): The new variable
#     """
#     return math.new_variable(value, bounds, name, dtype)


# def new_constant(value, name: str, dtype=math.float64) -> Tensor:
#     r"""Returns a new constant (non-trainable) tensor from the current math backend
#     with initial value set by `value`.
#     Args:
#         value (numeric): The initial value of the tensor
#         name (str): The name of the constant
#         dtype: The dtype of the constant
#
#     Returns:
#         tensor (Tensor): The new constant tensor
#     """
#     return math.new_constant(value, name, dtype)


def new_symplectic(num_modes: int) -> Tensor:
    r"""Returns a new symplectic matrix from the current math backend with ``num_modes`` modes.

    Args:
        num_modes (int): the number of modes in the symplectic matrix

    Returns:
        Tensor: the new symplectic matrix
    """
    return math.random_symplectic(num_modes)


def new_orthogonal(num_modes: int) -> Tensor:
    return math.random_orthogonal(num_modes)


def update_symplectic(
    symplectic_params: Sequence[Trainable], symplectic_grads: Sequence[Tensor], symplectic_lr: float
):

    r"""Updates the symplectic parameters using the given symplectic gradients.

    Implemented from:
        Wang J, Sun H, Fiori S. A Riemannian‐steepest‐descent approach
        for optimization on the real symplectic group.
        Mathematical Methods in the Applied Sciences. 2018 Jul 30;41(11):4273-86.
    """
    for S, dS_euclidean in zip(symplectic_params, symplectic_grads):
        Y = math.euclidean_to_symplectic(S, dS_euclidean)
        YT = math.transpose(Y)
        new_value = math.matmul(
            S, math.expm(-symplectic_lr * YT) @ math.expm(-symplectic_lr * (Y - YT))
        )
        math.assign(S, new_value)


def update_orthogonal(
    orthogonal_params: Sequence[Trainable], orthogonal_grads: Sequence[Tensor], orthogonal_lr: float
):
    r"""Updates the orthogonal parameters using the given orthogonal gradients.

    Implemented from:
        Fiori S, Bengio Y. Quasi-Geodesic Neural Learning Algorithms
        Over the Orthogonal Group: A Tutorial.
        Journal of Machine Learning Research. 2005 May 1;6(5).
    """
    for O, dO_euclidean in zip(orthogonal_params, orthogonal_grads):
        dO_orthogonal = 0.5 * (
            dO_euclidean - math.matmul(math.matmul(O, math.transpose(dO_euclidean)), O)
        )
        new_value = math.matmul(
            O, math.expm(orthogonal_lr * math.matmul(math.transpose(dO_orthogonal), O))
        )
        math.assign(O, new_value)


def update_euclidean(
    euclidean_params: Sequence[Trainable], euclidean_grads: Sequence[Tensor], euclidean_lr: float
):
    math.euclidean_opt.lr = euclidean_lr
    math.euclidean_opt.apply_gradients(zip(euclidean_grads, euclidean_params))



# this is the topmost __init__.py file of the mrmustard package

# from rich.pretty import install  # NOTE: just for the looks
# install()
from ._version import __version__


class Settings:
    """Settings class."""

    def __init__(self):
        self._backend = "tensorflow"
        self.HBAR = 2.0
        self.CHOI_R = 0.881373587019543  # np.arcsinh(1.0)
        self.DEBUG = False
        # clip(mean + 5*std, min, max) when auto-detecting the Fock cutoff
        self.AUTOCUTOFF_STDEV_FACTOR = 5
        self.AUTOCUTOFF_MAX_CUTOFF = 100
        self.AUTOCUTOFF_MIN_CUTOFF = 1
        # using cutoff=5 for each mode when determining if two transformations in fock repr are equal
        self.EQ_TRANSFORMATION_CUTOFF = 5
        self.EQ_TRANSFORMATION_RTOL_FOCK = 1e-3
        self.EQ_TRANSFORMATION_RTOL_GAUSS = 1e-6
        # for the detectors
        self.PNR_INTERNAL_CUTOFF = 50
        self.HOMODYNE_SQUEEZING = 10.0

    @property
    def backend(self):
        """The backend which is used.

        Can be either ``'tensorflow'`` or ``'torch'``.
        """
        return self._backend

    @backend.setter
    def backend(self, backend_name: str):
        if backend_name not in ["tensorflow", "torch"]:
            raise ValueError("Backend must be either 'tensorflow' or 'torch'")
        self._backend = backend_name


settings = Settings()
"""Settings object."""

settings.backend = "tensorflow"


def version():
    r"""Version number of Mr Mustard.

    Returns:
      str: package version number
    """
    return __version__


def about():
    """Mr Mustard information.

    Prints the installed version numbers for Mr Mustard and its dependencies,
    and some system info. Please include this information in bug reports.

    **Example:**

    .. code-block:: pycon

        >>> mm.about()
        Mr Mustard: a differentiable bridge between phase space and Fock space.
        Copyright 2018-2021 Xanadu Quantum Technologies Inc.

        Python version:            3.6.10
        Platform info:             Linux-5.8.18-1-MANJARO-x86_64-with-arch-Manjaro-Linux
        Installation path:         /home/mrmustard/
        Mr Mustard version:        0.1.0
        Numpy version:             1.21.4
        Numba version:             0.48.0
        Scipy version:             1.7.3
        The Walrus version:        0.17.0
        TensorFlow version:        2.7.0
        Torch version:             1.10.0+cu102
    """
    # pylint: disable=import-outside-toplevel
    import sys
    import platform
    import os
    import numpy
    import numba
    import scipy
    import thewalrus
    import tensorflow

    # a QuTiP-style infobox
    print("\nMr Mustard: a differentiable bridge between phase space and Fock space.")
    print("Copyright 2018-2021 Xanadu Quantum Technologies Inc.\n")

    print("Python version:            {}.{}.{}".format(*sys.version_info[0:3]))
    print("Platform info:             {}".format(platform.platform()))
    print("Installation path:         {}".format(os.path.dirname(__file__)))
    print("Mr Mustard version:        {}".format(__version__))
    print("Numpy version:             {}".format(numpy.__version__))
    print("Numba version:             {}".format(numba.__version__))
    print("Scipy version:             {}".format(scipy.__version__))
    print("The Walrus version:        {}".format(thewalrus.__version__))
    print("TensorFlow version:        {}".format(tensorflow.__version__))

    try:
        import torch

        torch_version = torch.__version__
    except ImportError:
        torch_version = None

    print("Torch version:             {}".format(torch_version))



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from mrmustard.types import *
from mrmustard import settings
from mrmustard.math import Math

math = Math()


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~ static functions ~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def fock_state(n: Sequence[int]) -> Tensor:
    r"""Returns a pure or mixed Fock state.

    Args:
        n: a list of photon numbers

    Returns:
        the Fock state up to cutoffs ``n+1``
    """
    psi = np.zeros(np.array(n) + np.ones_like(n), dtype=np.complex128)
    psi[tuple(np.atleast_1d(n))] = 1
    return psi


def autocutoffs(
    number_stdev: Matrix, number_means: Vector, max_cutoff: int = None, min_cutoff: int = None
) -> Tuple[int, ...]:
    r"""Returns the autocutoffs of a Wigner state.

    Args:
        number_stdev: the photon number standard deviation in each mode
            (i.e. the square root of the diagonal of the covariance matrix)
        number_means: the photon number means vector
        max_cutoff: the maximum cutoff

    Returns:
        Tuple[int, ...]: the suggested cutoffs
    """
    if max_cutoff is None:
        max_cutoff = settings.AUTOCUTOFF_MAX_CUTOFF
    if min_cutoff is None:
        min_cutoff = settings.AUTOCUTOFF_MIN_CUTOFF
    autocutoffs = settings.AUTOCUTOFF_MIN_CUTOFF + math.cast(
        number_means + number_stdev * settings.AUTOCUTOFF_STDEV_FACTOR, "int32"
    )
    return [int(n) for n in math.clip(autocutoffs, min_cutoff, max_cutoff)]


def fock_representation(
    cov: Matrix,
    means: Vector,
    shape: Sequence[int],
    return_dm: bool = None,
    return_unitary: bool = None,
    choi_r: float = None,
) -> Tensor:
    r"""Returns the Fock representation of a state or Choi state.

    * If the state is pure it returns the state vector (ket).
    * If the state is mixed it returns the density matrix.
    * If the transformation is unitary it returns the unitary transformation matrix.
    * If the transformation is not unitary it returns the Choi matrix.

    Args:
        cov: the Wigner covariance matrix
        means: the Wigner means vector
        shape: the shape of the tensor
        return_dm: whether the state vector is mixed or not
        return_unitary: whether the transformation is unitary or not
        choi_r: the TMSV squeezing magnitude

    Returns:
        Tensor: the fock representation
    """
    if return_dm is not None and return_unitary is not None:
        raise ValueError("Cannot specify both mixed and unitary.")
    if return_dm is None and return_unitary is None:
        raise ValueError("Must specify either mixed or unitary.")
    if return_unitary is not None and choi_r is None:
        raise ValueError("Must specify the choi_r value.")
    if return_dm is not None:  # i.e. it's a state
        A, B, C = ABC(cov, means, full=return_dm)
    elif return_unitary is not None and choi_r is not None:  # i.e. it's a transformation
        A, B, C = ABC(cov, means, full=not return_unitary, choi_r=choi_r)
    return math.hermite_renormalized(
        math.conj(-A), math.conj(B), math.conj(C), shape=shape
    )  # NOTE: remove conj when TW is updated


def ket_to_dm(ket: Tensor) -> Tensor:
    r"""Maps a ket to a density matrix.

    Args:
        ket: the ket

    Returns:
        Tensor: the density matrix
    """
    return math.outer(ket, math.conj(ket))


def ket_to_probs(ket: Tensor) -> Tensor:
    r"""Maps a ket to probabilities.

    Args:
        ket: the ket

    Returns:
        Tensor: the probabilities vector
    """
    return math.abs(ket) ** 2


def dm_to_probs(dm: Tensor) -> Tensor:
    r"""Extracts the diagonals of a density matrix.

    Args:
        dm: the density matrix

    Returns:
        Tensor: the probabilities vector
    """
    return math.all_diagonals(dm, real=True)


def U_to_choi(U: Tensor) -> Tensor:
    r"""Converts a unitary transformation to a Choi tensor.

    Args:
        U: the unitary transformation

    Returns:
        Tensor: the Choi tensor
    """
    cutoffs = U.shape[: len(U.shape) // 2]
    N = len(cutoffs)
    outer = math.outer(U, math.conj(U))
    return math.transpose(
        outer,
        list(range(0, N))
        + list(range(2 * N, 3 * N))
        + list(range(N, 2 * N))
        + list(range(3 * N, 4 * N)),
    )  # NOTE: mode blocks 1 and 3 are at the end so we can tensordot dm with them


def ABC(cov, means, full: bool, choi_r: float = None) -> Tuple[Matrix, Vector, Scalar]:
    r"""Returns the full-size ``A`` matrix, ``B`` vector and ``C`` scalar.

    Args:
        cov: the Wigner covariance matrix
        means: the Wigner means vector
        full: whether to return the full-size ``A``, ``B`` and ``C`` or the half-size ``A``, ``B``
            and ``C``
        choi_r: the TMSV squeezing magnitude if not None we consider ABC of a Choi state

    Returns:
        Tuple[Matrix, Vector, Scalar]: full-size ``A`` matrix, ``B`` vector and ``C`` scalar
    """
    is_state = choi_r is None
    N = cov.shape[-1] // 2
    R = math.rotmat(N)
    sigma = math.matmul(math.matmul(R, cov / settings.HBAR), math.dagger(R))
    beta = math.matvec(R, means / math.sqrt(settings.HBAR, dtype=means.dtype))
    Q = sigma + 0.5 * math.eye(2 * N, dtype=sigma.dtype)  # Husimi covariance matrix
    Qinv = math.inv(Q)
    A = math.matmul(math.Xmat(N), math.eye(2 * N, dtype=Qinv.dtype) - Qinv)
    denom = math.sqrt(math.det(Q)) if is_state else math.sqrt(math.det(Q / np.cosh(choi_r)))
    if full:
        B = math.matvec(math.transpose(Qinv), math.conj(beta))
        exponent = -0.5 * math.sum(math.conj(beta)[:, None] * Qinv * beta[None, :])
        C = math.exp(exponent) / denom
    else:
        A = A[
            :N, :N
        ]  # TODO: find a way to compute the half-size A without computing the full-size A first
        B = beta[N:] - math.matvec(A, beta[:N])
        exponent = -0.5 * math.sum(beta[:N] * B)
        C = math.exp(exponent) / math.sqrt(denom)
    if choi_r is not None:
        ones = math.ones(
            N // 2, dtype=A.dtype
        )  # N//2 is the actual number of modes because of the choi trick
        factor = 1.0 / np.tanh(choi_r)
        if full:
            rescaling = math.concat([ones, factor * ones, ones, factor * ones], axis=0)
        else:
            rescaling = math.concat([ones, factor * ones], axis=0)
        A = rescaling[:, None] * rescaling[None, :] * A
        B = rescaling * B
    return A, B, C


def fidelity(state_a, state_b, a_ket: bool, b_ket: bool) -> Scalar:
    r"""Computes the fidelity between two states in Fock representation."""
    if a_ket and b_ket:
        min_cutoffs = tuple([slice(min(a, b)) for a, b in zip(state_a.shape, state_b.shape)])
        state_a = state_a[min_cutoffs]
        state_b = state_b[min_cutoffs]
        return math.abs(math.sum(math.conj(state_a) * state_b)) ** 2
    elif a_ket:
        min_cutoffs = tuple(
            [
                slice(min(a, b))
                for a, b in zip(state_a.shape, state_b.shape[: len(state_b.shape) // 2])
            ]
        )
        state_a = state_a[min_cutoffs]
        state_b = state_b[min_cutoffs * 2]
        a = math.reshape(state_a, -1)
        return math.real(
            math.sum(math.conj(a) * math.matvec(math.reshape(state_b, (len(a), len(a))), a))
        )
    elif b_ket:
        min_cutoffs = tuple(
            [
                slice(min(a, b))
                for a, b in zip(state_a.shape[: len(state_a.shape) // 2], state_b.shape)
            ]
        )
        state_a = state_a[min_cutoffs * 2]
        state_b = state_b[min_cutoffs]
        b = math.reshape(state_b, -1)
        return math.real(
            math.sum(math.conj(b) * math.matvec(math.reshape(state_a, (len(b), len(b))), b))
        )
    else:
        raise NotImplementedError("Fidelity between mixed states is not implemented yet.")


def number_means(tensor, is_dm: bool):
    r"""Returns the mean of the number operator in each mode."""
    probs = math.all_diagonals(tensor, real=True) if is_dm else math.abs(tensor) ** 2
    modes = [m for m in range(len(probs.shape))]
    marginals = [math.sum(probs, axes=modes[:k] + modes[k + 1 :]) for k in range(len(modes))]
    return math.astensor(
        [
            math.sum(marginal * math.arange(len(marginal), dtype=marginal.dtype))
            for marginal in marginals
        ]
    )


def number_variances(tensor, is_dm: bool):
    r"""Returns the variance of the number operator in each mode."""
    probs = math.all_diagonals(tensor, real=True) if is_dm else math.abs(tensor) ** 2
    modes = [m for m in range(len(probs.shape))]
    marginals = [math.sum(probs, axes=modes[:k] + modes[k + 1 :]) for k in range(len(modes))]
    return math.astensor(
        [
            (
                math.sum(marginal * math.arange(marginal.shape[0], dtype=marginal.dtype) ** 2)
                - math.sum(marginal * math.arange(marginal.shape[0], dtype=marginal.dtype)) ** 2
            )
            for marginal in marginals
        ]
    )


def purity(dm: Tensor) -> Scalar:
    r"""Returns the purity of a density matrix."""
    cutoffs = dm.shape[: len(dm.shape) // 2]
    d = int(np.prod(cutoffs))  # combined cutoffs in all modes
    dm = math.reshape(dm, (d, d))
    dm = normalize(dm, is_dm=True)
    return math.abs(math.sum(math.transpose(dm) * dm))  # tr(rho^2)


def CPTP(transformation, fock_state, transformation_is_unitary: bool, state_is_dm: bool) -> Tensor:
    r"""Computes the CPTP (note: CP, really) channel given by a transformation (unitary matrix or choi operator) on a state.

    It assumes that the cutoffs of the transformation matches the cutoffs of the relevant axes of the state.

    Args:
        transformation: the transformation tensor
        fock_state: the state to transform
        transformation_is_unitary: whether the transformation is a unitary matrix or a Choi operator
        state_is_dm: whether the state is a density matrix or a ket

    Returns:
        Tensor: the transformed state
    """
    num_modes = len(fock_state.shape) // 2 if state_is_dm else len(fock_state.shape)
    N0 = list(range(0, num_modes))
    N1 = list(range(num_modes, 2 * num_modes))
    N2 = list(range(2 * num_modes, 3 * num_modes))
    N3 = list(range(3 * num_modes, 4 * num_modes))
    if transformation_is_unitary:
        U = transformation
        Us = math.tensordot(U, fock_state, axes=(N1, N0))
        if not state_is_dm:
            return Us
        else:  # is state is dm, the input indices of dm are still at the end of Us
            return math.tensordot(Us, math.dagger(U), axes=(N1, N0))
    else:
        C = transformation  # choi operator
        if state_is_dm:
            return math.tensordot(C, fock_state, axes=(N1 + N3, N0 + N1))
        else:
            Cs = math.tensordot(C, fock_state, axes=(N1, N0))
            return math.tensordot(
                Cs, math.conj(fock_state), axes=(N2, N0)
            )  # N2 is the last set of indices now


def contract_states(
    stateA, stateB, a_is_mixed: bool, b_is_mixed: bool, modes: List[int], normalize: bool
):
    r"""Contracts two states in the specified modes, it assumes that the modes spanned by ``B`` are a subset of the modes spanned by ``A``.

    Args:
        stateA: the first state
        stateB: the second state (assumed to be on a subset of the modes of stateA)
        a_is_mixed: whether the first state is mixed or not.
        b_is_mixed: whether the second state is mixed or not.
        modes: the modes on which to contract the states.
        normalize: whether to normalize the result

    Returns:
        State: the contracted state (subsystem of ``A``)
    """
    indices = list(range(len(modes)))
    if not a_is_mixed and not b_is_mixed:
        out = math.tensordot(math.conj(stateB), stateA, axes=(indices, modes))
        if normalize:
            out = out / math.norm(out)
        return out
    elif a_is_mixed and not b_is_mixed:
        Ab = math.tensordot(
            stateA, stateB, axes=([m + len(stateA.shape) // 2 for m in modes], indices)
        )
        out = math.tensordot(math.conj(stateB), Ab, axes=(indices, modes))
    elif not a_is_mixed and b_is_mixed:
        Ba = math.tensordot(stateB, stateA, axes=(indices, modes))  # now B indices are all first
        out = math.tensordot(math.conj(stateA), Ba, axes=(modes, indices))
    elif a_is_mixed and b_is_mixed:
        out = math.tensordot(
            stateA,
            math.conj(stateB),
            axes=(
                list(modes) + [m + len(stateA.shape) // 2 for m in modes],
                list(indices) + [i + len(stateB.shape) // 2 for i in indices],
            ),
        )
    if normalize:
        out = out / math.sum(math.all_diagonals(out, real=False))
    return out


def normalize(fock: Tensor, is_dm: bool):
    if is_dm:
        return fock / math.sum(math.all_diagonals(fock, real=False))
    else:
        return fock / math.sum(math.norm(fock))


def norm(state: Tensor, is_dm: bool):
    r"""
    Returns the norm of a ket or the trace of the density matrix.
    Note that the "norm" is intended as the float number that is used to normalize the state,
    and depends on the representation. Hence different numbers for different representations
    of the same state (:math:`|amp|` for ``ket`` and :math:`|amp|^2` for ``dm``).
    """
    if is_dm:
        return math.sum(math.all_diagonals(state, real=True))
    else:
        return math.abs(math.norm(state))


def is_mixed_dm(dm):
    cutoffs = dm.shape[: len(dm.shape) // 2]
    square = math.reshape(dm, (int(np.prod(cutoffs)), -1))
    return not np.isclose(math.sum(square * math.transpose(square)), 1.0)


def trace(dm, keep: List[int]):
    r"""Computes the partial trace of a density matrix.

    Args:
        dm: the density matrix
        keep: the modes to keep
    """
    N = len(dm.shape) // 2
    trace = [m for m in range(N) if m not in keep]
    # put at the end all of the indices to trace over
    dm = math.transpose(
        dm, [i for pair in [(k, k + N) for k in keep] + [(t, t + N) for t in trace] for i in pair]
    )
    d = int(np.prod(dm.shape[-len(trace) :]))
    # make it square on those indices
    dm = math.reshape(dm, dm.shape[: 2 * len(keep)] + (d, d))
    return math.trace(dm)



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The lab module is all you need to construct and simulate photonic circuits.
It contains the items you'd find in a lab:

* states (Vacuum, Coherent, SqueezedVacuum, Thermal, etc.)
* transformations (Sgate, BSgate, LossChannel, etc.)
* detectors (PNRDetector, Homodyne, etc.)
* the Circuit class
"""

from .circuit import *
from .states import *
from .gates import *
from .detectors import *
from .abstract import State



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import *

# NOTE: when type-annotating with typevars, objects with the same typevars must have the same type
# E.g. in `def f(x: Vector, y: Vector) -> Tensor: ...`
# the type of `x` and the type of `y` are assumed to be the same, even though "Vector" can mean different things.
Scalar = TypeVar("Scalar")
Vector = TypeVar("Vector")
Matrix = TypeVar("Matrix")
Tensor = TypeVar("Tensor")
Array = TypeVar("Array")  # TODO: let mypy know that this is Vector, Matrix, or Tensor
Trainable = TypeVar("Trainable")

Numeric = Union[Scalar, Vector, Matrix, Tensor]



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from mrmustard.types import *
from abc import ABC, abstractmethod, abstractproperty
from itertools import product
import numpy as np
from mrmustard.math import Math

math = Math()


class XPTensor(ABC):
    r"""A representation of Matrices and Vectors in phase space.

    Tensors in phase space have a ``(2n, 2n)`` or ``(2n,)`` shape where n is the number of modes.

    There are two main orderings:
        - xxpp: matrix is a `2\times 2` block matrix where each block is an `xx`, `xp`, `px`, `pp` block on all modes.
        - xpxp: matrix is a `n\times n` block matrix of `2\times 2` blocks each corresponding to a mode or a coherence between modes
    This creates some difficulties when we need to work in a mode-wise fashion, especially whith coherences.
    We solve this problem by reshaping the matrices to `(n,m,2,2)` and vectors to `(n,2)`.

    We call `n` the outmodes and `m` the inmodes.
    Off-diagonal matrices like coherences have all the outmodes different than the inmodes.
    Diagonal matrices like coviariances and symplectic transformations have the same outmodes as the inmodes.
    Vectors have only outmodes.

    XPTensor objects are sparse, in the sense that they support implicit operations between modes where one or more tensors are undefined.
    There are two types of behaviour:
        * like_0 (default): in modes where the tensor is undefined, it's like having a zero (a zero matrix)
        * like_1: in modes where the tensor is undefined, it's like having a one (an identity matrix)

    For example, in the expression :math:`X @ means + d` where `X` is a symplectic matrix and `d` is a displacement vector,
    if `X` is undefined it's like having the identity and the matrix product simply returns means, while in the expression
    `means + d` if `d` is undefined it simply returns `means`. In this example no operation was actually computed.
    Thanks to sparsity we can represent graph states and transformations on graph states as XPTensor objects.

    Args:
        tensor: The tensor in (n,m,2,2) or (n,2) order.
        modes: a list of modes for a diagonal matrix or a vector and a tuple of two lists for a coherence (not optional for a coherence)
        like_0: Whether the null tensor behaves like 0 under addition (e.g. the noise matrix Y)
        like_1: Whether the null tensor behaves like 1 under multiplication (e.g. a symplectic transformation matrix)
    """

    @abstractmethod  # so that XPTensor can't be instantiated directly
    def __init__(
        self,
        tensor: Optional[Tensor],
        like_0: bool,
        isVector: bool,
        modes: Tuple[List[int], List[int]],
    ):

        self.like_0 = like_0
        self.shape = (
            None if tensor is None else tensor.shape[: len(tensor.shape) // 2]
        )  # only (N,M) or (N,)
        self.ndim = None if tensor is None else len(self.shape)
        self.isVector = isVector
        if self.ndim == 1 and not self.isVector:
            raise ValueError(
                f"tensor shape incompatible with isVector={isVector} (expected len(tensor.shape)==4, got {len(tensor.shape)})"
            )
        if self.ndim == 2 and self.isVector:
            raise ValueError(
                f"tensor shape incompatible with isVector={isVector} (expected len(tensor.shape)==2, got {len(tensor.shape)})"
            )
        if self.isVector and self.like_1:
            raise ValueError("vectors should be like_0")
        self.tensor = tensor
        if not (set(modes[0]) == set(modes[1]) or set(modes[0]).isdisjoint(modes[1])):
            raise ValueError(
                "The inmodes and outmodes should either contain the same modes or be disjoint"
            )
        self.modes = modes

    @property
    def dtype(self):
        return None if self.tensor is None else self.tensor.dtype

    @property
    def outmodes(self) -> List[int]:
        return self.modes[0]

    @property
    def inmodes(self) -> List[int]:
        return self.modes[1]

    @property
    def num_modes(self) -> int:
        # TODO: raise warning for coherence blocks?
        return len(self.outmodes)

    @property
    def isMatrix(self) -> Optional[bool]:
        return not self.isVector

    @property
    def isCoherence(self) -> Optional[bool]:
        return self.isMatrix and self.outmodes != self.inmodes

    @property
    def like_1(self) -> bool:
        return not self.like_0

    @property
    def T(self) -> XPMatrix:
        if self.isVector:
            raise ValueError("Cannot transpose a vector")
        if self.tensor is None:
            return self
        return XPMatrix(
            math.transpose(self.tensor, (1, 0, 3, 2)),
            self.like_0,
            self.like_1,
            (self.inmodes, self.outmodes),
        )

    def to_xpxp(self) -> Optional[Union[Matrix, Vector]]:
        if self.tensor is None:
            return None
        tensor = math.transpose(
            self.tensor, (0, 2, 1, 3) if self.isMatrix else (0, 1)
        )  # from NN22 to N2N2 or from N2 to N2
        return math.reshape(tensor, [2 * s for s in self.shape])

    def to_xxpp(self) -> Optional[Union[Matrix, Vector]]:
        if self.tensor is None:
            return None
        tensor = math.transpose(
            self.tensor, (2, 0, 3, 1) if self.isMatrix else (1, 0)
        )  # from NN22 to 2N2N or from N2 to 2N
        return math.reshape(tensor, [2 * s for s in self.shape])

    def __array__(self):
        return self.to_xxpp()

    def modes_first(self) -> Optional[Tensor]:
        return self.tensor

    def modes_last(self) -> Optional[Tensor]:
        if self.tensor is None:
            return None
        return math.transpose(self.tensor, (2, 3, 0, 1) if self.isMatrix else (0, 1))  # 22NM or 2N

    def clone(self, times: int, modes=None) -> XPtensor:
        r"""Create a new XPTensor made by cloning the system a given number of times
        (the modes are reset by default unless specified).
        """
        if self.tensor is None:
            return self
        if times == 1:
            return self
        if self.isMatrix:
            tensor = math.expand_dims(self.modes_last(), axis=4)  # shape = [2,2,N,N,1]
            tensor = math.tile(tensor, (1, 1, 1, 1, times))  # shape = [2,2,N,N,T]
            tensor = math.diag(tensor)  # shape = [2,2,N,N,T,T]
            tensor = math.transpose(tensor, (0, 1, 2, 4, 3, 5))  # shape = [2,2,N,T,N,T]
            tensor = math.reshape(
                tensor, (2, 2, tensor.shape[2] * times, tensor.shape[4] * times)
            )  # shape = [2,2,NT,NT] = [2,2,O,O]
            tensor = math.transpose(tensor, (2, 3, 0, 1))  # shape = [NT,NT,2,2]
            return XPMatrix(tensor, self.like_0, self.like_1, ([], []) if modes is None else modes)
        else:
            tensor = math.tile(
                self.expand_dims(self.modes_last(), axis=2), (1, 1, times)
            )  # shape = [2,N,T]
            tensor = math.reshape(tensor, (2, -1))  # shape = [2,NT] = [2,O]
            tensor = math.transpose(tensor, (1, 0))  # shape = [NT,2] = [O,2]
            return XPVector(tensor, [] if modes is None else modes)

    def clone_like(self, other: XPTensor):
        r"""Create a new XPTensor with the same shape and modes as other.

        The new tensor has the same content as self, cloned as many times as necessary to match the
        shape and modes of other. The other properties are kept as is.

        Args:
            other: The tensor to be cloned.

        Returns:
            Tensor: A new XPTensor with the same shape and modes as other.
        """
        if other.shape == self.shape:
            return self
        if self.isCoherence:
            raise ValueError("Cannot clone a coherence block")
        if bool(other.num_modes % self.num_modes):
            raise ValueError(
                f"No integer multiple of {self.num_modes} modes fits into {other.num_modes} modes"
            )
        times = other.num_modes // self.num_modes
        if self.isVector == other.isVector:
            tensor = self.clone(times, modes=other.modes).tensor
        else:
            raise ValueError("Cannot clone a vector into a matrix or viceversa")
        if self.isMatrix:
            return XPMatrix(tensor, self.like_0, self.like_1, (other.outmodes, other.inmodes))
        else:
            return XPVector(tensor, other.outmodes)

    ####################################################################################################################
    # Operators
    ####################################################################################################################

    def __rmul__(self, other: Scalar) -> XPTensor:
        "Implements the operation ``self * other``"
        if self.tensor is None:
            if self.like_1:
                raise NotImplementedError("Cannot multiply a scalar and a like_1 null tensor yet")
            else:
                return self
        self.tensor = other * self.tensor
        return self

    def __mul__(self, other: Scalar) -> Optional[XPTensor]:
        return other * self

    def __matmul__(self, other: Union[XPMatrix, XPVector]) -> Union[XPMatrix, XPVector, Scalar]:
        if not isinstance(other, (XPMatrix, XPVector)):
            raise TypeError(
                f"Unsupported operand type(s) for @: '{self.__class__.__qualname__}' and '{other.__class__.__qualname__}'"
            )
        # TODO: move mode-check at beginning?

        # both are None
        if self.tensor is None and other.tensor is None:
            if self.isMatrix and other.isMatrix:
                return XPMatrix(None, like_1=self.like_1 and other.like_1)
            elif self.isVector or other.isVector:
                return XPVector(None)
        # either is None
        if self.tensor is None:
            return self if self.like_0 else other
        if other.tensor is None:
            return other if other.like_0 else self
        # Now neither self nor other is None
        if self.isMatrix and other.isMatrix:
            tensor, modes = self._mode_aware_matmul(other)
            return XPMatrix(tensor, like_1=self.like_1 and other.like_1, modes=modes)
        elif self.isMatrix and other.isVector:
            tensor, modes = self._mode_aware_matmul(other)
            return XPVector(
                tensor, modes[0]
            )  # TODO: check if we can output modes as a list in _mode_aware_matmul
        elif self.isVector and other.isMatrix:
            tensor, modes = other.T._mode_aware_matmul(self)
            return XPVector(tensor, modes[0])
        else:  # self.isVector and other.isVector:
            return self._mode_aware_vecvec(other)  # NOTE: this is a scalar, not an XPTensor

    def _mode_aware_matmul(
        self, other: Union[XPMatrix, XPVector]
    ) -> Tuple[Tensor, Tuple[List[int], List[int]]]:
        r"""Performs matrix multiplication only on the necessary modes and
        takes care of keeping only the modes that are needed, in case of mismatch.

        See documentation for a visual explanation with blocks.  #TODO: add link to figure
        """
        if list(self.inmodes) == list(other.outmodes):  # NOTE: they match including the ordering
            prod = math.tensordot(
                self.tensor, other.tensor, ((1, 3), (0, 2)) if other.isMatrix else ((1, 3), (0, 1))
            )
            return math.transpose(prod, (0, 2, 1, 3) if other.isMatrix else (0, 1)), (
                self.outmodes,
                other.inmodes,
            )
        contracted = [i for i in self.inmodes if i in other.outmodes]
        uncontracted_self = [i for i in self.inmodes if i not in contracted]
        uncontracted_other = [o for o in other.outmodes if o not in contracted]
        if not (
            set(self.outmodes).isdisjoint(uncontracted_other)
            and set(other.inmodes).isdisjoint(uncontracted_self)
        ):
            raise ValueError("Invalid modes")
        bulk = None
        copied_rows = None
        copied_cols = None
        if len(contracted) > 0:
            subtensor1 = math.gather(
                self.tensor, [self.inmodes.index(m) for m in contracted], axis=1
            )
            subtensor2 = math.gather(
                other.tensor, [other.outmodes.index(m) for m in contracted], axis=0
            )
            if other.isMatrix:
                bulk = math.tensordot(subtensor1, subtensor2, ((1, 3), (0, 2)))
                bulk = math.transpose(bulk, (0, 2, 1, 3))
            else:
                bulk = math.tensordot(subtensor1, subtensor2, ((1, 3), (0, 1)))
        if self.like_1 and len(uncontracted_other) > 0:
            copied_rows = math.gather(
                other.tensor, [other.outmodes.index(m) for m in uncontracted_other], axis=0
            )
        if other.like_1 and len(uncontracted_self) > 0:
            copied_cols = math.gather(
                self.tensor, [self.inmodes.index(m) for m in uncontracted_self], axis=1
            )
        if copied_rows is not None and copied_cols is not None:
            if bulk is None:
                bulk = math.zeros(
                    (copied_cols.shape[0], copied_rows.shape[1], 2, 2), dtype=copied_cols.dtype
                )
            empty = math.zeros(
                (copied_rows.shape[0], copied_cols.shape[1], 2, 2), dtype=copied_cols.dtype
            )
            final = math.block([[copied_cols, bulk], [empty, copied_rows]], axes=[0, 1])
        elif copied_cols is None and copied_rows is not None:
            if bulk is None:
                final = copied_rows
            else:
                final = math.block([[bulk], [copied_rows]], axes=[0, 1])
        elif copied_rows is None and copied_cols is not None:
            if bulk is None:
                final = copied_cols
            else:
                final = math.block([[copied_cols, bulk]], axes=[0, 1])
        else:  # copied_rows and copied_cols are both None
            final = bulk  # NOTE: could be None

        outmodes = self.outmodes + uncontracted_other
        if other.like_0 and len(contracted) == 0:
            outmodes = uncontracted_other
        if self.like_0:
            outmodes = [m for m in outmodes if m in self.outmodes]

        inmodes = uncontracted_self + other.inmodes
        if self.like_0 and len(contracted) == 0:
            inmodes = uncontracted_self
        if other.like_0:
            inmodes = [m for m in inmodes if m in other.inmodes]

        if final is not None:
            final = math.gather(final, [outmodes.index(o) for o in sorted(outmodes)], axis=0)
            if other.isMatrix:
                final = math.gather(final, [inmodes.index(i) for i in sorted(inmodes)], axis=1)
        return final, (sorted(outmodes), sorted(inmodes))

    def _mode_aware_vecvec(self, other: XPVector) -> Scalar:
        if list(self.outmodes) == list(other.outmodes):
            return math.sum(self.tensor * other.tensor)
        common = list(
            set(self.outmodes) & set(other.outmodes)
        )  # only the common modes (the others are like 0)
        return math.sum(self.tensor[common] * other.tensor[common])

    def __add__(self, other: Union[XPMatrix, XPVector]) -> Union[XPMatrix, XPVector]:
        if not isinstance(other, (XPMatrix, XPVector)):
            raise TypeError(
                f"unsupported operand type(s) for +: '{self.__class__.__qualname__}' and '{other.__class__.__qualname__}'"
            )
        if self.isVector != other.isVector:
            raise ValueError("Cannot add a vector and a matrix")
        if self.isCoherence != other.isCoherence:
            raise ValueError("Cannot add a coherence block and a diagonal block")
        if self.tensor is None and other.tensor is None:  # both are None
            if self.like_1 and other.like_1:
                raise ValueError("Cannot add two like_1 null tensors yet")  # because 1+1 = 2
            if self.isMatrix and other.isMatrix:
                return XPMatrix(like_0=self.like_0 and other.like_0)
            else:
                return XPVector()
        if self.tensor is None:  # only self is None
            if self.like_0:
                return other
            elif (
                self.like_1
            ):  # other must be a matrix because self is like_1, so it must be a matrix and we can't add a vector to a matrix
                indices = [
                    [i, i] for i in range(other.num_modes)
                ]  # TODO: check if this is always correct
                updates = math.tile(
                    math.expand_dims(math.eye(2, dtype=other.dtype), 0), (other.num_modes, 1, 1)
                )
                other.tensor = math.update_add_tensor(other.tensor, indices, updates)
                return other
        if other.tensor is None:  # only other is None
            return other + self
        # now neither is None
        modes_match = list(self.outmodes) == list(other.outmodes) and list(self.inmodes) == list(
            other.inmodes
        )
        if modes_match:
            self.tensor = self.tensor + other.tensor
            return self
        if not modes_match and self.like_1 and other.like_1:
            raise ValueError("Cannot add two like_1 tensors on different modes yet")
        outmodes = sorted(set(self.outmodes).union(other.outmodes))
        inmodes = sorted(set(self.inmodes).union(other.inmodes))
        self_contains_other = set(self.outmodes).issuperset(other.outmodes) and set(
            self.inmodes
        ).issuperset(other.inmodes)
        other_contains_self = set(other.outmodes).issuperset(self.outmodes) and set(
            other.inmodes
        ).issuperset(self.inmodes)
        if self_contains_other:
            to_update = self.tensor
            to_add = [other]
        elif other_contains_self:
            to_update = other.tensor
            to_add = [self]
        else:  # need to add both to a new empty tensor
            to_update = math.zeros(
                (len(outmodes), len(inmodes), 2, 2) if self.isMatrix else (len(outmodes), 2),
                dtype=self.tensor.dtype,
            )
            to_add = [self, other]
        for t in to_add:
            outmodes_indices = [outmodes.index(o) for o in t.outmodes]
            inmodes_indices = [inmodes.index(i) for i in t.inmodes]
            if (
                t.isMatrix
            ):  # e.g. outmodes of to_update are [self]+[other_new] = (e.g.) [9,1,2]+[0,20]
                indices = [[o, i] for o in outmodes_indices for i in inmodes_indices]
            else:
                indices = [[o] for o in outmodes_indices]
            to_update = math.update_add_tensor(
                to_update,
                indices,
                math.reshape(t.modes_first(), (-1, 2, 2) if self.isMatrix else (-1, 2)),
            )
        if self.isMatrix and other.isMatrix:
            return XPMatrix(
                to_update,
                like_0=self.like_0 and other.like_0,
                like_1=self.like_1 or other.like_1,
                modes=(outmodes, inmodes),
            )
        else:
            return XPVector(to_update, outmodes)

    def __sub__(self, other: Union[XPMatrix, XPVector]) -> Optional[XPTensor]:
        return self + (-1) * other

    def __truediv__(self, other: Scalar) -> Optional[XPTensor]:
        return (1 / other) * self

    def __getitem__(self, modes: Union[int, slice, List[int], Tuple]) -> Union[XPMatrix, XPVector]:
        r"""Returns modes or subsets of modes from the XPTensor or coherences between modes using an
        intuitive notation.

        We handle mode indices and we get the corresponding tensor indices handled correctly.

        Examples:

        .. code::

            T[N] ~ self.tensor[N,:,:,:]
            T[M,N] = the coherence between the modes M and N
            T[:,N] ~ self.tensor[:,N,:,:]
            T[[1,2,3],:] ~ self.tensor[[1,2,3],:,:,:] # i.e. the block with outmodes [1,2,3] and all inmodes
            T[[1,2,3],[4,5]] ~ self.tensor[[1,2,3],[4,5],:,:]  # i.e. the block with outmodes [1,2,3] and inmodes [4,5]
        """
        if self.isVector:
            if isinstance(modes, int):
                _modes = [modes]
            elif isinstance(modes, list) and all(isinstance(m, int) for m in modes):
                _modes = modes
            elif modes == slice(None, None, None):
                _modes = self.outmodes
            else:
                raise ValueError(f"Usage: V[1], V[[1,2,3]] or V[:]")
            rows = [self.outmodes.index(m) for m in modes]
            return XPVector(math.gather(self.tensor, rows, axis=0), modes)
        else:
            _modes = [None, None]
            if isinstance(modes, int):
                _modes = ([modes], slice(None, None, None))
            elif isinstance(modes, list) and all(isinstance(m, int) for m in modes):
                _modes = (modes, slice(None, None, None))
            elif modes == slice(None, None, None):
                _modes = (slice(None, None, None), slice(None, None, None))
            elif isinstance(modes, tuple) and len(modes) == 2:
                for i, M in enumerate(modes):
                    if isinstance(M, int):
                        _modes[i] = [M]
                    elif isinstance(M, list):
                        _modes[i] = M
                    elif M == slice(None, None, None):
                        _modes[i] = self.modes[i]
                    else:
                        raise ValueError(
                            f"Invalid modes: {M} from {modes} (tensor has modes {self.modes})"
                        )
            else:
                raise ValueError(f"Invalid modes: {modes} (tensor has modes {self.modes})")
            rows = [self.outmodes.index(m) for m in _modes[0]]
            columns = [self.inmodes.index(m) for m in _modes[1]]
            subtensor = math.gather(self.tensor, rows, axis=0)
            subtensor = math.gather(subtensor, columns, axis=1)
            return XPMatrix(
                subtensor,
                like_1=_modes[0] == _modes[1] if self.like_1 else False,
                modes=tuple(_modes),
            )


class XPMatrix(XPTensor):
    r"""A convenience class for a matrix in the XPTensor format.

    # TODO: write docstring
    """

    def __init__(
        self,
        tensor: Tensor = None,
        like_0: bool = None,
        like_1: bool = None,
        modes: Tuple[List[int], List[int]] = ([], []),
    ):
        if like_0 is None and like_1 is None:
            raise ValueError("At least one of like_0 or like_1 must be set")
        if like_0 == like_1:
            raise ValueError(f"like_0 and like_1 can't both be {like_0}")
        if not (
            isinstance(modes, tuple) and len(modes) == 2 and all(type(m) == list for m in modes)
        ):
            raise ValueError("modes should be a tuple containing two lists (outmodes and inmodes)")
        if len(modes[0]) == 0 and len(modes[1]) == 0 and tensor is not None:
            if (
                tensor.shape[0] != tensor.shape[1] and like_0
            ):  # NOTE: we can't catch square coherences if no modes are specified
                raise ValueError("Must specify the modes for a coherence block")
            modes = tuple(
                list(range(s)) for s in tensor.shape[:2]
            )  # NOTE assuming that it isn't a coherence block
        like_0 = like_0 if like_0 is not None else not like_1
        super().__init__(tensor, like_0, isVector=False, modes=modes)

    @classmethod
    def from_xxpp(
        cls,
        tensor: Optional[Union[Matrix, Vector]],
        like_0: Optional[bool] = None,
        like_1: Optional[bool] = None,
        modes: Tuple[List[int], List[int]] = ([], []),
    ) -> XPMatrix:
        if tensor is not None:
            tensor = math.reshape(tensor, [_ for n in tensor.shape for _ in (2, n // 2)])
            tensor = math.transpose(tensor, (1, 3, 0, 2))
        return XPMatrix(tensor, like_0, like_1, modes)

    @classmethod
    def from_xpxp(
        cls,
        tensor: Optional[Union[Matrix, Vector]],
        like_0: bool = None,
        like_1: bool = None,
        modes: Tuple[List[int], List[int]] = ([], []),
    ) -> XPMatrix:
        if tensor is not None:
            tensor = math.reshape(tensor, [_ for n in tensor.shape for _ in (n // 2, 2)])
            tensor = math.transpose(tensor, (0, 2, 1, 3))
        return XPMatrix(tensor, like_0, like_1, modes)

    def __repr__(self) -> str:
        return f"XPMatrix(like_0={self.like_0}, modes={self.modes}, tensor_xpxp=\n{self.to_xpxp()})"


class XPVector(XPTensor):
    r"""A convenience class for a vector in the XPTensor format."""

    def __init__(self, tensor: Tensor = None, modes: List[int] = []):
        if not (isinstance(modes, list) or all(type(m) == int for m in modes)):
            raise ValueError(f"the modes of an XPVector should be a list of ints")
        if len(modes) == 0 and tensor is not None:
            modes = list(range(tensor.shape[0]))
        super().__init__(tensor, like_0=True, isVector=True, modes=(modes, []))

    @classmethod
    def from_xxpp(
        cls,
        tensor: Optional[Union[Matrix, Vector]],
        modes: List[int] = [],
    ) -> XPMatrix:
        if tensor is not None:
            tensor = math.reshape(tensor, (2, -1))
            tensor = math.transpose(tensor, (1, 0))
        return XPVector(tensor, modes)

    @classmethod
    def from_xpxp(
        cls,
        tensor: Optional[Union[Matrix, Vector]],
        modes: List[int] = [],
    ) -> XPMatrix:
        if tensor is not None:
            tensor = math.reshape(tensor, (-1, 2))
        return XPVector(tensor, modes)

    def __repr__(self) -> str:
        return f"XPVector(modes={self.outmodes}, tensor_xpxp=\n{self.to_xpxp()})"



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from setuptools import setup, find_packages

with open("mrmustard/_version.py") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")

requirements = [
    "numpy",
    "scipy",
    "numba",
    "thewalrus>=0.17.0",
    "tensorflow>=2.4.0",
    "rich",
    "tqdm",
    "matplotlib",
]

info = {
    "name": "mrmustard",
    "version": version,
    "description": "Differentiable quantum Gaussian circuits",
    "url": "https://github.com/XanaduAI/mrmustard",
    "author": "Xanadu",
    "author_email": "filippo@xanadu.ai",
    "license": "Apache License 2.0",
    "packages": find_packages(where="."),
    "install_requires": requirements,
    "long_description": open("README.md", encoding="utf-8").read(),
    "long_description_content_type": "text/markdown",
}

classifiers = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: Apache Software License",
    "Natural Language :: English",
    "Operating System :: POSIX",
    "Operating System :: MacOS :: MacOS X",
    "Operating System :: POSIX :: Linux",
    "Operating System :: Microsoft :: Windows",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3 :: Only",
    "Topic :: Scientific/Engineering :: Physics",
]

setup(classifiers=classifiers, **(info))



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from abc import ABC, abstractmethod
import numpy as np
from rich.table import Table
from rich import print as rprint

from mrmustard.physics import gaussian, fock
from mrmustard.lab.abstract.state import State
from mrmustard.types import *
from mrmustard.utils import graphics
from mrmustard.utils.parametrized import Parametrized
from mrmustard import settings
from mrmustard.math import Math

math = Math()


class Transformation:
    r"""Base class for all Transformations."""
    _bell = None  # single-mode TMSV state for gaussian-to-fock conversion
    is_unitary = True  # whether the transformation is unitary (True by default)

    def primal(self, state: State) -> State:
        r"""Applies ``self`` (a ``Transformation``) to other (a ``State``) and returns the transformed state.

        Args:
            state (State): the state to transform

        Returns:
            State: the transformed state
        """
        if state.is_gaussian:
            new_state = self.transform_gaussian(state, dual=False)
        else:
            new_state = self.transform_fock(state, dual=False)
        return new_state

    def dual(self, state: State) -> State:
        r"""Applies the dual of self (dual of a ``Transformation``) to other (a ``State``) and returns the transformed state.

        Args:
            state (State): the state to transform

        Returns:
            State: the transformed state
        """
        if state.is_gaussian:
            new_state = self.transform_gaussian(state, dual=True)
        else:
            new_state = self.transform_fock(state, dual=True)
        return new_state

    @property
    def bell(self):
        r"""The N-mode two-mode squeezed vacuum for the choi-jamiolkowksi isomorphism."""
        if self._bell is None:
            cov = gaussian.two_mode_squeezed_vacuum_cov(
                r=settings.CHOI_R, phi=0.0, hbar=settings.HBAR
            )
            means = gaussian.vacuum_means(num_modes=2, hbar=settings.HBAR)
            bell = bell_single = State(cov=cov, means=means)
            for _ in range(self.num_modes - 1):
                bell = bell & bell_single
            tot = 2 * self.num_modes
            order = tuple(range(0, tot, 2)) + tuple(range(1, tot, 2))
            self._bell = bell.get_modes(order)
        return self._bell[self.modes + [m + self.num_modes for m in self.modes]]

    def transform_gaussian(self, state: State, dual: bool) -> State:
        r"""Transforms a Gaussian state into a Gaussian state.

        Args:
            state (State): the state to transform
            dual (bool): whether to apply the dual channel

        Returns:
            State: the transformed state
        """
        X, Y, d = self.XYd if not dual else self.XYd_dual
        cov, means = gaussian.CPTP(state.cov, state.means, X, Y, d, state.modes, self.modes)
        new_state = State(
            cov=cov, means=means, modes=state.modes, _norm=state._norm
        )  # NOTE: assumes modes don't change
        return new_state

    def transform_fock(self, state: State, dual: bool) -> State:
        r"""Transforms a state in Fock representation.

        Args:
            state (State): the state to transform
            dual (bool): whether to apply the dual channel

        Returns:
            State: the transformed state
        """
        if self.is_unitary:
            U = self.U(cutoffs=state.cutoffs)
            transformation = fock.math.dagger(U) if dual else U
        else:
            transformation = self.choi(cutoffs=state.cutoffs)
            if dual:
                n = len(state.cutoffs)
                N0 = list(range(0, n))
                N1 = list(range(n, 2 * n))
                N2 = list(range(2 * n, 3 * n))
                N3 = list(range(3 * n, 4 * n))
                transformation = fock.math.transpose(transformation, N3 + N0 + N1 + N2)
        new_fock = fock.CPTP(
            transformation=transformation,
            fock_state=state.ket(state.cutoffs) if state.is_pure else state.dm(state.cutoffs),
            transformation_is_unitary=self.is_unitary,
            state_is_dm=state.is_mixed,
        )
        if state.is_mixed or not self.is_unitary:
            return State(dm=new_fock, modes=state.modes)
        else:
            return State(ket=new_fock, modes=state.modes)

    @property
    def modes(self) -> Sequence[int]:
        if self._modes in (None, []):
            X, Y, d = self.XYd
            if d is not None:
                self._modes = list(range(d.shape[-1] // 2))
            elif X is not None:
                self._modes = list(range(X.shape[-1] // 2))
            elif Y is not None:
                self._modes = list(range(Y.shape[-1] // 2))
        return self._modes

    @modes.setter
    def modes(self, modes: List[int]):
        r"""Sets the modes on which the transformation acts."""
        self._validate_modes(modes)
        self._modes = modes

    @property
    def num_modes(self) -> int:
        r"""The number of modes on which the transformation acts."""
        return len(self.modes)

    def _validate_modes(self, modes):
        pass

    @property
    def X_matrix(self) -> Optional[Matrix]:
        return None

    @property
    def Y_matrix(self) -> Optional[Matrix]:
        return None

    @property
    def d_vector(self) -> Optional[Vector]:
        return None

    @property
    def X_matrix_dual(self) -> Optional[Matrix]:
        if (X := self.X_matrix) is None:
            return None
        return gaussian.math.inv(X)

    @property
    def Y_matrix_dual(self) -> Optional[Matrix]:
        if (Y := self.Y_matrix) is None:
            return None
        elif (Xdual := self.X_matrix_dual) is None:
            return Y
        else:
            return math.matmul(math.matmul(Xdual, Y), math.transpose(Xdual))

    @property
    def d_vector_dual(self) -> Optional[Vector]:
        if (d := self.d_vector) is None:
            return None
        elif (Xdual := self.X_matrix_dual) is None:
            return d
        else:
            return math.matmul(Xdual, d)

    @property
    def XYd(self) -> Tuple[Optional[Matrix], Optional[Matrix], Optional[Vector]]:
        r"""Returns the ```(X, Y, d)``` triple.

        Override in subclasses if computing ``X``, ``Y`` and ``d`` together is more efficient.
        """
        return self.X_matrix, self.Y_matrix, self.d_vector

    @property
    def XYd_dual(self) -> Tuple[Optional[Matrix], Optional[Matrix], Optional[Vector]]:
        r"""Returns the ```(X, Y, d)``` triple of the dual of the current transformation.

        Override in subclasses if computing ``Xdual``, ``Ydual`` and ``ddual`` together is more efficient.
        """
        return self.X_matrix_dual, self.Y_matrix_dual, self.d_vector_dual

    @property
    def is_phase_covariant(self) -> bool:
        X, Y, d = self.XYd
        if d is not None:
            return False
        if X is not None and not math.allclose(self.X, math.diag(math.diag_part(self.X))):
            return False
        if Y is not None and not math.allclose(self.Y, math.diag(math.diag_part(self.Y))):
            return False
        return True

    def U(self, cutoffs: Sequence[int]):
        r"""Returns the unitary representation of the transformation."""
        if not self.is_unitary:
            return None
        choi_state = self.bell >> self
        return fock.fock_representation(
            choi_state.cov,
            choi_state.means,
            shape=cutoffs * 2,
            return_unitary=True,
            choi_r=settings.CHOI_R,
        )

    def choi(self, cutoffs: Sequence[int]):
        r"""Returns the Choi representation of the transformation."""
        if self.is_unitary:
            U = self.U(cutoffs)
            return fock.U_to_choi(U)
        else:
            choi_state = self.bell >> self
            choi_op = fock.fock_representation(
                choi_state.cov,
                choi_state.means,
                shape=cutoffs * 4,
                return_unitary=False,
                choi_r=settings.CHOI_R,
            )
            return choi_op

    def __getitem__(self, items) -> Callable:
        r"""Sets the modes on which the transformation acts.

        Allows transformations to be used as: ``output = transf[0,1](input)``,  e.g. acting on
        modes 0 and 1.
        """
        #  TODO: this won't work when we want to reuse the same op for different modes in a circuit.
        # i.e. `psi = op[0](psi); psi = op[1](psi)` is ok, but `circ = Circuit([op[0], op[1]])` won't work.
        if isinstance(items, int):
            modes = [items]
        elif isinstance(items, slice):
            modes = list(range(items.start, items.stop, items.step))
        elif isinstance(items, (Sequence, Iterable)):
            modes = list(items)
        else:
            raise ValueError(f"{items} is not a valid slice or list of modes.")
        self.modes = modes
        return self

    # TODO: use __class_getitem__ for compiler stuff

    def __rshift__(self, other: Transformation):
        r"""Concatenates self with other (other after self).

        If any of the two is a circuit, all the ops in it migrate to the new circuit that is returned.
        E.g., ``circ = Sgate(1.0)[0,1] >> Dgate(0.2)[0] >> BSgate(np.pi/4)[0,1]``

        Args:
            other: another transformation

        Returns:
            Circuit: A circuit that concatenates self with other
        """
        from mrmustard.lab import Circuit  # this is called at runtime so it's ok

        ops1 = self._ops if isinstance(self, Circuit) else [self]
        ops2 = other._ops if isinstance(other, Circuit) else [other]
        return Circuit(ops1 + ops2)

    def __lshift__(self, other: Union[State, Transformation]):
        r"""Applies the dual of self to other.

        If other is a state, the dual of self is applied to the state.
        If other is a transformation, the dual of self is concatenated after other (in the dual sense).

        E.g.
        .. code-block::

            Sgate(0.1) << Coherent(0.5)   # state
            Sgate(0.1) << Dgate(0.2)      # transformation

        Args:
            other: a state or a transformation

        Returns:
            State: the state transformed via the dual transformation or the transformation
            concatenated after other
        """
        if isinstance(other, State):
            return self.dual(other)
        elif isinstance(other, Transformation):
            return self >> other  # so that the dual is self.dual(other.dual(x))
        else:
            raise ValueError(f"{other} is not a valid state or transformation.")

    def __eq__(self, other):
        r"""Returns ``True`` if the two transformations are equal."""
        if not isinstance(other, Transformation):
            return False
        if self.is_gaussian and other.is_gaussian:
            sX, sY, sd = self.XYd
            oX, oY, od = other.XYd
            if sX is None:
                if oX is not None:
                    if not np.allclose(
                        oX, np.eye(oX.shape[0]), rtol=settings.EQ_TRANSFORMATION_RTOL_GAUSS
                    ):
                        return False
            if oX is None:
                if sX is not None:
                    if not np.allclose(
                        sX, np.eye(sX.shape[0]), rtol=settings.EQ_TRANSFORMATION_RTOL_GAUSS
                    ):
                        return False
            if sX is not None and oX is not None:
                if not np.allclose(sX, oX):
                    return False
            if sY is None:
                if oY is not None:
                    if not np.allclose(
                        oY, np.zeros_like(oY), rtol=settings.EQ_TRANSFORMATION_RTOL_GAUSS
                    ):
                        return False
            if oY is None:
                if sY is not None:
                    if not np.allclose(
                        sY, np.zeros_like(sY), rtol=settings.EQ_TRANSFORMATION_RTOL_GAUSS
                    ):
                        return False
            if sY is not None and oY is not None:
                if not np.allclose(sY, oY):
                    return False
            if sd is None:
                if od is not None:
                    if not np.allclose(
                        sd, np.zeros_like(sd), rtol=settings.EQ_TRANSFORMATION_RTOL_GAUSS
                    ):
                        return False
            if od is None:
                if sd is not None:
                    if not np.allclose(
                        sd, np.zeros_like(sd), rtol=settings.EQ_TRANSFORMATION_RTOL_GAUSS
                    ):
                        return False
            if sd is not None and od is not None:
                if not np.allclose(sd, od):
                    return False
            return True
        else:
            return np.allclose(
                self.choi(cutoffs=[settings.EQ_TRANSFORMATION_CUTOFF] * self.num_modes),
                other.choi(cutoffs=[settings.EQ_TRANSFORMATION_CUTOFF] * self.num_modes),
                rtol=settings.EQ_TRANSFORMATION_RTOL_FOCK,
            )

    def __repr__(self):
        table = Table(title=f"{self.__class__.__qualname__} on modes {self.modes}")
        table.add_column("Parameters")
        table.add_column("dtype")
        table.add_column("Value")
        table.add_column("Bounds")
        table.add_column("Shape")
        table.add_column("Trainable")
        with np.printoptions(precision=6, suppress=True):
            for name in self._param_names:
                par = self.__dict__[name]
                table.add_row(
                    name,
                    par.dtype.name,
                    f"{np.array(par)}",
                    str(self.__dict__["_" + name + "_bounds"]),
                    f"{par.shape}",
                    str(self.__dict__["_" + name + "_trainable"]),
                )
        rprint(table)
        return ""



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Optional, List, Tuple, Sequence, Dict
from mrmustard.types import Scalar, Vector, Matrix, Trainable
from mrmustard import settings
from mrmustard.lab.abstract import State
from mrmustard.physics import gaussian, fock
from mrmustard.utils.parametrized import Parametrized
from mrmustard.utils import training
from mrmustard.math import Math

math = Math()


__all__ = [
    "Vacuum",
    "SqueezedVacuum",
    "Coherent",
    "Thermal",
    "DisplacedSqueezed",
    "TMSV",
    "Gaussian",
    "Fock",
]


class Vacuum(State):
    r"""The N-mode vacuum state."""

    def __init__(self, num_modes: int):
        cov = gaussian.vacuum_cov(num_modes, settings.HBAR)
        means = gaussian.vacuum_means(num_modes, settings.HBAR)
        State.__init__(self, cov=cov, means=means)


class Coherent(Parametrized, State):
    r"""The N-mode coherent state.

    Equivalent to applying a displacement to the vacuum state:

    .. code-block::

        Coherent(x=0.5, y=0.2) == Vacuum(1) >> Dgate(x=0.5, y=0.3)    # True

    Parallelizable over x and y:

    .. code-block::

        Coherent(x=[1.0, 2.0], y=[-1.0, -2.0]) == Coherent(x=1.0, y=-1.0) & Coherent(x=2.0, y=-2.0)  # True

    Can be used to model a heterodyne detection:

    .. code-block::

        Gaussian(2) << Coherent(x=1.0, y=0.0)[1]  # e.g. heterodyne on mode 1, leftover state on mode 0

    Note that the values of x and y are automatically rescaled by 1/(2*sqrt(mrmustard.settings.HBAR)).

    Args:
        x (float or List[float]): The x-displacement of the coherent state.
        y (float or List[float]): The y-displacement of the coherent state.
        x_trainable (bool): Whether the x-displacement is trainable.
        y_trainable (bool): Whether the y-displacement is trainable.
        x_bounds (float or None, float or None): The bounds of the x-displacement.
        y_bounds (float or None, float or None): The bounds of the y-displacement.
        modes (optional List[int]): The modes of the coherent state.
        cutoffs (Sequence[int], default=None): set to force the cutoff dimensions of the state
        normalize (bool, default False): whether to normalize the leftover state when projecting onto ``Coherent``
    """

    def __init__(
        self,
        x: Union[Optional[float], Optional[List[float]]] = 0.0,
        y: Union[Optional[float], Optional[List[float]]] = 0.0,
        x_trainable: bool = False,
        y_trainable: bool = False,
        x_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        y_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[Sequence[int]] = None,
        cutoffs: Optional[Sequence[int]] = None,
        normalize: bool = False,
    ):
        Parametrized.__init__(
            self,
            x=x,
            y=y,
            x_trainable=x_trainable,
            y_trainable=y_trainable,
            x_bounds=x_bounds,
            y_bounds=y_bounds,
            modes=modes,
            normalize=normalize,
        )
        means = gaussian.displacement(self.x, self.y, settings.HBAR)
        cov = gaussian.vacuum_cov(means.shape[-1] // 2, settings.HBAR)
        State.__init__(self, cov=cov, means=means, cutoffs=cutoffs)

    @property
    def means(self):
        return gaussian.displacement(self.x, self.y, settings.HBAR)


class SqueezedVacuum(Parametrized, State):
    r"""The N-mode squeezed vacuum state.

    Equivalent to applying a squeezing gate to the vacuum state:

    .. code::

      >>> SqueezedVacuum(r=0.5, phi=0.2) == Vacuum(1) >> Sgate(r=0.5, phi=0.2)
      True

    Parallelizable over r and phi:
    .. code::

      >>> SqueezedVacuum(r=[1.0, 2.0], phi=[-1.0, -2.0]) == SqueezedVacuum(r=1.0, phi=-1.0) & SqueezedVacuum(r=2.0, phi=-2.0)
      True

    Can be used to model a heterodyne detection with result 0.0:
    .. code::

      >>> Gaussian(2) << SqueezedVacuum(r=10.0, phi=0.0)[1]  # e.g. homodyne on x quadrature on mode 1 with result 0.0
      # leftover state on mode 0

    Args:
        r (float): the squeezing magnitude
        phi (float): The squeezing phase.
        r_trainable (bool): Whether the squeezing magnitude is trainable.
        phi_trainable (bool): Whether the squeezing phase is trainable.
        r_bounds (tuple): The bounds of the squeezing magnitude.
        phi_bounds (tuple): The bounds of the squeezing phase.
        modes (list): The modes of the squeezed vacuum state.
        cutoffs (Sequence[int], default=None): set to force the cutoff dimensions of the state
        normalize (bool, default False): whether to normalize the leftover state when projecting onto ``SqueezedVacuum``,
    """

    def __init__(
        self,
        r: Union[Scalar, Vector] = 0.0,
        phi: Union[Scalar, Vector] = 0.0,
        r_trainable: bool = False,
        phi_trainable: bool = False,
        r_bounds: Tuple[Optional[float], Optional[float]] = (0, None),
        phi_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[Sequence[int]] = None,
        cutoffs: Optional[Sequence[int]] = None,
        normalize: bool = False,
    ):
        Parametrized.__init__(
            self,
            r=r,
            phi=phi,
            r_trainable=r_trainable,
            phi_trainable=phi_trainable,
            r_bounds=r_bounds,
            phi_bounds=phi_bounds,
            modes=modes,
            normalize=normalize,
        )
        cov = gaussian.squeezed_vacuum_cov(self.r, self.phi, settings.HBAR)
        means = gaussian.vacuum_means(cov.shape[-1] // 2, settings.HBAR)
        State.__init__(self, cov=cov, means=means, cutoffs=cutoffs)

    @property
    def cov(self):
        return gaussian.squeezed_vacuum_cov(self.r, self.phi, settings.HBAR)


class TMSV(Parametrized, State):
    r"""The 2-mode squeezed vacuum state.

    Equivalent to applying a 50/50 beam splitter to a pair of squeezed vacuum states:

    .. code::

      >>> TMSV(r=0.5, phi=0.0) == Vacuum(2) >> Sgate(r=[0.5,0.5], phi=[0.0, np.pi]) >> BSgate(theta=-np.pi/4)
      True

    Args:
        r (float): The squeezing magnitude.
        phi (float): The squeezing phase.
        r_trainable (bool): Whether the squeezing magnitude is trainable.
        phi_trainable (bool): Whether the squeezing phase is trainable.
        r_bounds (tuple): The bounds of the squeezing magnitude.
        phi_bounds (tuple): The bounds of the squeezing phase.
        modes (list): The modes of the two-mode squeezed vacuum state. Must be of length 2.
        cutoffs (Sequence[int], default=None): set to force the cutoff dimensions of the state
        normalize (bool, default False): whether to normalize the leftover state when projecting onto ``TMSV``
    """

    def __init__(
        self,
        r: Union[Scalar, Vector] = 0.0,
        phi: Union[Scalar, Vector] = 0.0,
        r_trainable: bool = False,
        phi_trainable: bool = False,
        r_bounds: Tuple[Optional[float], Optional[float]] = (0, None),
        phi_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[Sequence[int]] = [0, 1],
        cutoffs: Optional[Sequence[int]] = None,
        normalize: bool = False,
    ):
        Parametrized.__init__(
            self,
            r=r,
            phi=phi,
            r_trainable=r_trainable,
            phi_trainable=phi_trainable,
            r_bounds=r_bounds,
            phi_bounds=phi_bounds,
            modes=modes,
            normalize=normalize,
        )
        cov = gaussian.two_mode_squeezed_vacuum_cov(self.r, self.phi, settings.HBAR)
        means = gaussian.vacuum_means(2, settings.HBAR)
        State.__init__(self, cov=cov, means=means, cutoffs=cutoffs)

    @property
    def cov(self):
        return gaussian.two_mode_squeezed_vacuum_cov(self.r, self.phi, settings.HBAR)


class Thermal(Parametrized, State):
    r"""The N-mode thermal state.

    Equivalent to applying additive noise to the vacuum:

    .. code::

        >>> Thermal(nbar=0.31) == Vacuum(1) >> AdditiveNoise(0.62)  # i.e. 2*nbar + 1 (from vac) in total
        True

    Parallelizable over ``nbar``:

    .. code::

        >>> Thermal(nbar=[0.1, 0.2]) == Thermal(nbar=0.1) & Thermal(nbar=0.2)
        True

    Args:
        nbar (float or List[float]): the expected number of photons in each mode
        nbar_trainable (bool): whether the ``nbar`` is trainable
        nbar_bounds (tuple): the bounds of the ``nbar``
        modes (list): the modes of the thermal state
        cutoffs (Sequence[int], default=None): set to force the cutoff dimensions of the state
        normalize (bool, default False): whether to normalize the leftover state when projecting onto ``Thermal``
    """

    def __init__(
        self,
        nbar: Union[Scalar, Vector] = 0.0,
        nbar_trainable: bool = False,
        nbar_bounds: Tuple[Optional[float], Optional[float]] = (0, None),
        modes: Optional[Sequence[int]] = None,
        cutoffs: Optional[Sequence[int]] = None,
        normalize: bool = False,
    ):
        Parametrized.__init__(
            self,
            nbar=nbar,
            nbar_trainable=nbar_trainable,
            nbar_bounds=nbar_bounds,
            modes=modes,
            normalize=normalize,
        )
        cov = gaussian.thermal_cov(self.nbar, settings.HBAR)
        means = gaussian.vacuum_means(cov.shape[-1] // 2, settings.HBAR)
        State.__init__(self, cov=cov, means=means, cutoffs=cutoffs)

    @property
    def cov(self):
        return gaussian.thermal_cov(self.nbar, settings.HBAR)


class DisplacedSqueezed(Parametrized, State):
    r"""The N-mode displaced squeezed state.

    Equivalent to applying a displacement to the squeezed vacuum state:

    .. code::

        >>> DisplacedSqueezed(r=0.5, phi=0.2, x=0.3, y=-0.7) == SqueezedVacuum(r=0.5, phi=0.2) >> Dgate(x=0.3, y=-0.7)
        True

    Parallelizable over ``r``, ``phi``, ``x``, ``y``:

    .. code::

        >>> DisplacedSqueezed(r=[0.1, 0.2], phi=[0.3, 0.4], x=[0.5, 0.6], y=[0.7, 0.8]) == DisplacedSqueezed(r=0.1, phi=0.3, x=0.5, y=0.7) & DisplacedSqueezed(r=0.2, phi=0.4, x=0.6, y=0.8)
        True

    Can be used to model homodyne detection:

    .. code::

      >>> Gaussian(2) << DisplacedSqueezed(r=10, phi=np.pi, y=0.3)[1]  # e.g. homodyne on mode 1, p quadrature, result 0.3
      # leftover state on mode 0

    Args:
        r (float or List[float]): the squeezing magnitude
        phi (float or List[float]): the squeezing phase
        x (float or List[float]): the displacement in the x direction
        y (float or List[float]): the displacement in the y direction
        r_trainable (bool): whether the squeezing magnitude is trainable
        phi_trainable (bool): whether the squeezing phase is trainable
        x_trainable (bool): whether the displacement in the x direction is trainable
        y_trainable (bool): whether the displacement in the y direction is trainable
        r_bounds (tuple): the bounds of the squeezing magnitude
        phi_bounds (tuple): the bounds of the squeezing phase
        x_bounds (tuple): the bounds of the displacement in the x direction
        y_bounds (tuple): the bounds of the displacement in the y direction
        modes (list): the modes of the displaced squeezed state.
        cutoffs (Sequence[int], default=None): set to force the cutoff dimensions of the state
        normalize (bool, default False): whether to normalize the leftover state when projecting onto ``DisplacedSqueezed``
    """

    def __init__(
        self,
        r: Union[Scalar, Vector] = 0.0,
        phi: Union[Scalar, Vector] = 0.0,
        x: Union[Scalar, Vector] = 0.0,
        y: Union[Scalar, Vector] = 0.0,
        r_trainable: bool = False,
        phi_trainable: bool = False,
        x_trainable: bool = False,
        y_trainable: bool = False,
        r_bounds: Tuple[Optional[float], Optional[float]] = (0, None),
        phi_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        x_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        y_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[Sequence[int]] = None,
        cutoffs: Optional[Sequence[int]] = None,
        normalize: bool = False,
    ):
        Parametrized.__init__(
            self,
            r=r,
            phi=phi,
            x=x,
            y=y,
            r_trainable=r_trainable,
            phi_trainable=phi_trainable,
            x_trainable=x_trainable,
            y_trainable=y_trainable,
            r_bounds=r_bounds,
            phi_bounds=phi_bounds,
            x_bounds=x_bounds,
            y_bounds=y_bounds,
            modes=modes,
            normalize=normalize,
        )
        cov = gaussian.squeezed_vacuum_cov(self.r, self.phi, settings.HBAR)
        means = gaussian.displacement(self.x, self.y, settings.HBAR)
        State.__init__(self, cov=cov, means=means, cutoffs=cutoffs)

    @property
    def cov(self):
        return gaussian.squeezed_vacuum_cov(self.r, self.phi, settings.HBAR)

    @property
    def means(self):
        return gaussian.displacement(self.x, self.y, settings.HBAR)


class Gaussian(Parametrized, State):
    r"""The N-mode Gaussian state parametrized by a symplectic matrix and N symplectic eigenvalues.

    The (mixed) Gaussian state is equivalent to applying a Gaussian symplectic transformation to a Thermal state:

    .. code::

        >>> G = Gaussian(num_modes=1, eigenvalues = np.random.uniform(settings.HBAR/2, 10.0))
        >>> G == Thermal(nbar=(G.eigenvalues*2/settings.HBAR  - 1)/2) >> Ggate(1, symplectic=G.symplectic)
        True

    Note that the 1st moments are zero unless a Dgate is applied to the Gaussian state:

    .. code::

        >>> np.allclose(Gaussian(num_modes=1).means, 0.0)
        True

    Args:
        num_modes (int): the number of modes
        eigenvalues (float or List[float]): the symplectic eigenvalues of the Gaussian state
        symplectic (np.ndarray or List[np.ndarray]): the symplectic matrix of the Gaussian state
        eigenvalues_trainable (bool): whether the eigenvalues are trainable
        symplectic_trainable (bool): whether the symplectic matrix is trainable
        eigenvalues_bounds (tuple): the bounds of the eigenvalues
        modes (optional, List[int]): the modes of the Gaussian state.
        cutoffs (Sequence[int], default=None): set to force the cutoff dimensions of the state
        normalize (bool, default False): whether to normalize the leftover state when projecting onto Gaussian
    """

    def __init__(
        self,
        num_modes: int,
        symplectic: Matrix = None,
        eigenvalues: Vector = None,
        symplectic_trainable: bool = False,
        eigenvalues_trainable: bool = False,
        eigenvalues_bounds: Tuple[Optional[float], Optional[float]] = (settings.HBAR / 2, None),
        modes: List[int] = None,
        cutoffs: Optional[Sequence[int]] = None,
        normalize: bool = False,
    ):
        if symplectic is None:
            symplectic = training.new_symplectic(num_modes=num_modes)
        if eigenvalues is None:
            eigenvalues = gaussian.math.ones(num_modes) * settings.HBAR / 2
        if math.any(math.atleast_1d(eigenvalues) < settings.HBAR / 2):
            raise ValueError(
                f"Eigenvalues cannot be smaller than hbar/2 = {settings.HBAR}/2 = {settings.HBAR/2}"
            )
        Parametrized.__init__(
            self,
            symplectic=symplectic,
            eigenvalues=eigenvalues,
            eigenvalues_trainable=eigenvalues_trainable,
            symplectic_trainable=symplectic_trainable,
            eigenvalues_bounds=eigenvalues_bounds,
            modes=modes,
            normalize=normalize,
        )
        cov = gaussian.gaussian_cov(self.symplectic, self.eigenvalues, settings.HBAR)
        means = gaussian.vacuum_means(cov.shape[-1] // 2, settings.HBAR)
        State.__init__(self, cov=cov, means=means, cutoffs=cutoffs)

    @property
    def cov(self):
        return gaussian.gaussian_cov(self.symplectic, self.eigenvalues, settings.HBAR)

    @property
    def is_mixed(self):
        return any(self.eigenvalues > settings.HBAR / 2)

    @property
    def trainable_parameters(self) -> Dict[str, List[Trainable]]:
        return {
            "symplectic": [self.symplectic] * self._symplectic_trainable,
            "orthogonal": [],
            "euclidean": ([self.eigenvalues] * self._eigenvalues_trainable),
        }


class Fock(Parametrized, State):
    r"""The N-mode Fock state.

    Args:
        n (int or List[int]): the number of photons in each mode
        modes (optional, List[int]): the modes of the Fock state
        cutoffs (Sequence[int], default=None): set to force the cutoff dimensions of the state
        normalize (bool, default False): whether to normalize the leftover state when projecting onto ``Fock``
    """

    def __init__(
        self,
        n: Sequence[int],
        modes: Sequence[int] = None,
        cutoffs: Sequence[int] = None,
        normalize: bool = False,
    ):
        State.__init__(self, ket=fock.fock_state(n), cutoffs=cutoffs)
        Parametrized.__init__(
            self, n=[n] if isinstance(n, int) else n, modes=modes, normalize=normalize
        )

    def _preferred_projection(self, other: State, mode_indices: Sequence[int]):
        r"""Preferred method to perform a projection onto this state (rather than the default one).

        E.g. ``ket << Fock(1, modes=[3])`` is equivalent to ``ket[:,:,:,1]`` if ``ket`` has 4 modes
        E.g. ``dm << Fock(1, modes=[1])`` is equivalent to ``dm[:,1,:,1]`` if ``dm`` has 2 modes

        Args:
            other: the state to project onto this state
            mode_indices: the indices of the modes of other that we want to project onto self
        """
        getitem = []
        used = 0
        for i, _ in enumerate(other.modes):
            if i in mode_indices:
                getitem.append(self._n[used])
                used += 1
            else:
                getitem.append(slice(None))
        output = other.fock[tuple(getitem)] if other.is_pure else other.fock[tuple(getitem) * 2]
        if self._normalize:
            return fock.normalize(output, is_dm=other.is_mixed)
        return output



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from mrmustard.types import *
from functools import lru_cache
from scipy.special import binom
from scipy.stats import unitary_group
from itertools import product
from functools import lru_cache, wraps


class Autocast:
    r"""A decorator that casts all castable arguments of a method to the dtype with highest precision."""

    def __init__(self):
        self.dtype_order = ("float16", "float32", "float64", "complex64", "complex128")
        self.no_cast = ("int8", "uint8", "int16", "uint16", "int32", "uint32", "int64", "uint64")

    def can_cast(self, arg):
        return hasattr(arg, "dtype") and arg.dtype.name not in self.no_cast

    def should_cast(self, arg, proposed_dtype):
        if not self.can_cast(arg):
            return False
        return self.dtype_order.index(proposed_dtype) > self.dtype_order.index(arg.dtype.name)

    def get_dtypes(self, *args, **kwargs) -> List:
        r"""Returns the dtypes of the arguments."""
        args_dtypes = [arg.dtype.name for arg in args if self.can_cast(arg)]
        kwargs_dtypes = [v.dtype.name for v in kwargs.values() if self.can_cast(v)]
        return args_dtypes + kwargs_dtypes

    def max_dtype(self, dtypes: List):
        r"""Returns the dtype with the highest precision."""
        if dtypes == []:
            return None
        return max(dtypes, key=lambda dtype: self.dtype_order.index(dtype))

    def cast_all(self, backend, *args, **kwargs):
        r"""Casts all arguments to the highest precision when possible and needed."""
        max_dtype = self.max_dtype(self.get_dtypes(*args, **kwargs))
        args = [
            backend.cast(arg, max_dtype) if self.should_cast(arg, max_dtype) else arg
            for arg in args
        ]
        kwargs = {
            k: backend.cast(v, max_dtype) if self.should_cast(v, max_dtype) else v
            for k, v in kwargs.items()
        }
        return args, kwargs

    def __call__(self, func):
        @wraps(func)
        def wrapper(backend, *args, **kwargs):
            args, kwargs = self.cast_all(backend, *args, **kwargs)
            return func(backend, *args, **kwargs)

        return wrapper



import pytest

from mrmustard import *
import numpy as np
import tensorflow as tf
from thewalrus.random import random_covariance
from thewalrus.quantum import real_to_complex_displacements
from mrmustard.physics import gaussian as gp


@pytest.mark.parametrize("hbar", [1 / 2, 1.0, 2.0, 1.6])
@pytest.mark.parametrize("num_modes", np.arange(5, 10))
@pytest.mark.parametrize("pure", [True, False])
@pytest.mark.parametrize("block_diag", [True, False])
def test_fidelity_is_symmetric(num_modes, hbar, pure, block_diag):
    """Test that the fidelity is symmetric and between 0 and 1"""
    cov1 = random_covariance(num_modes, hbar=hbar, pure=pure, block_diag=block_diag)
    means1 = np.sqrt(2 * hbar) * np.random.rand(2 * num_modes)
    cov2 = random_covariance(num_modes, hbar=hbar, pure=pure, block_diag=block_diag)
    means2 = np.sqrt(2 * hbar) * np.random.rand(2 * num_modes)
    f12 = gp.fidelity(means1, cov1, means2, cov2, hbar)
    f21 = gp.fidelity(means2, cov2, means1, cov1, hbar)
    assert np.allclose(f12, f21)
    assert 0 <= np.real_if_close(f12) < 1.0


@pytest.mark.parametrize("hbar", [1 / 2, 1.0, 2.0, 1.6])
@pytest.mark.parametrize("num_modes", np.arange(5, 10))
@pytest.mark.parametrize("pure", [True, False])
@pytest.mark.parametrize("block_diag", [True, False])
def test_fidelity_with_self(num_modes, hbar, pure, block_diag):
    """Test that the fidelity of two identical quantum states is 1"""
    cov = random_covariance(num_modes, hbar=hbar, pure=pure, block_diag=block_diag)
    means = np.random.rand(2 * num_modes)
    assert np.allclose(gp.fidelity(means, cov, means, cov, hbar=hbar), 1, atol=1e-4)


@pytest.mark.parametrize("num_modes", np.arange(5, 10))
@pytest.mark.parametrize("hbar", [0.5, 1.0, 2.0, 1.6])
def test_fidelity_coherent_state(num_modes, hbar):
    """Test the fidelity of two multimode coherent states"""
    beta1 = np.random.rand(num_modes) + 1j * np.random.rand(num_modes)
    beta2 = np.random.rand(num_modes) + 1j * np.random.rand(num_modes)
    means1 = real_to_complex_displacements(np.concatenate([beta1, beta1.conj()]), hbar=hbar)
    means2 = real_to_complex_displacements(np.concatenate([beta2, beta2.conj()]), hbar=hbar)
    cov1 = hbar * np.identity(2 * num_modes) / 2
    cov2 = hbar * np.identity(2 * num_modes) / 2
    fid = gp.fidelity(means1, cov1, means2, cov2, hbar=hbar)
    expected = np.exp(-np.linalg.norm(beta1 - beta2) ** 2)
    assert np.allclose(expected, fid)


@pytest.mark.parametrize("hbar", [0.5, 1.0, 2.0, 1.6])
@pytest.mark.parametrize("r1", np.random.rand(3))
@pytest.mark.parametrize("r2", np.random.rand(3))
def test_fidelity_squeezed_vacuum(r1, r2, hbar):
    """Tests fidelity between two squeezed states"""
    cov1 = np.diag([np.exp(2 * r1), np.exp(-2 * r1)]) * hbar / 2
    cov2 = np.diag([np.exp(2 * r2), np.exp(-2 * r2)]) * hbar / 2
    mu = np.zeros([2])
    assert np.allclose(1 / np.cosh(r1 - r2), gp.fidelity(mu, cov1, mu, cov2, hbar=hbar))


@pytest.mark.parametrize("n1", [0.5, 1.0, 2.0, 1.6])
@pytest.mark.parametrize("n2", [0.5, 1.0, 2.0, 1.6])
@pytest.mark.parametrize("hbar", [0.5, 1.0, 2.0, 1.6])
def test_fidelity_thermal(n1, n2, hbar):
    """Test fidelity between two thermal states"""
    expected = 1 / (1 + n1 + n2 + 2 * n1 * n2 - 2 * np.sqrt(n1 * n2 * (n1 + 1) * (n2 + 1)))
    cov1 = hbar * (n1 + 0.5) * np.identity(2)
    cov2 = hbar * (n2 + 0.5) * np.identity(2)
    mu1 = np.zeros([2])
    mu2 = np.zeros([2])
    assert np.allclose(expected, gp.fidelity(mu1, cov1, mu2, cov2, hbar=hbar))


@pytest.mark.parametrize("hbar", [0.5, 1.0, 2.0, 1.6])
@pytest.mark.parametrize("r", [-2.0, 0.0, 2.0])
@pytest.mark.parametrize("alpha", np.random.rand(10) + 1j * np.random.rand(10))
def test_fidelity_vac_to_displaced_squeezed(r, alpha, hbar):
    """Calculates the fidelity between a coherent squeezed state and vacuum"""
    cov1 = np.diag([np.exp(2 * r), np.exp(-2 * r)]) * hbar / 2
    means1 = real_to_complex_displacements(np.array([alpha, np.conj(alpha)]), hbar=hbar)
    means2 = np.zeros([2])
    cov2 = np.identity(2) * hbar / 2
    expected = (
        np.exp(-np.abs(alpha) ** 2) * np.abs(np.exp(np.tanh(r) * np.conj(alpha) ** 2)) / np.cosh(r)
    )
    assert np.allclose(expected, gp.fidelity(means1, cov1, means2, cov2, hbar=hbar))
